import tkinter as tk
import tkinter.simpledialog as tksimpledialog

from tkinter import ttk

from itertools import accumulate
import bisect

from sol.config import GlobalConfig
C = GlobalConfig()

class ClipControl:
    def __init__(self, root, backend, layer):

        self.root = root
        self.backend = backend

        # important savedata
        self.width = 330
        self.layer = layer
        self.cur_pos = 0.0
        self.cur_clip = None

        self.pad_buts = []
        self.cue_but_states = []
        self.loop_but_states = []
        self.lp_data = None
        self.pad_but_cmds = []

        self.deleting_mode = False

        # all tk vars
        self.name_var = tk.StringVar()
        self.zoom_follow_var = tk.BooleanVar()

        self.sens_var, self.xtra_sens_var = tk.DoubleVar(), tk.StringVar()
        self.speed_var, self.xtra_speed_var = tk.DoubleVar(), tk.StringVar()

        self.loop_selected_text_var = tk.StringVar()

        # these are overriden upon creation
        self.qp_lp_var = None
        self.loop_on_var = None
        self.loop_type_var = None

        self.loop_on_toggle = None
        self.loop_type_switch = None

        # clip parameter to update function
        self.param_dispatch = {
            'cue_points': self.update_cues,
            'loop_points': self.update_loop,
            'loop_on': self.update_loop_on,
            'loop_type': self.update_loop_type,
            'loop_selection': self.update_loop,
            'playback_speed': self.update_speed,
            'control_sens': self.update_sens
        }

        # direct update fun
        self.direct_updates = {
            'pad_activate': self.activate_pad,
            'pad_deactivate': self.delet_pad,
            'pads_toggle': self.toggle_pad_mode,
            'pad_press': self.press_pad,
            'qp_pad_press': self.press_pad_qp,
            'lp_pad_press': self.press_pad_lp,
        }

        base_addr = '/magi/layer{}'.format(self.layer)
        bfn = self.backend.fun_store

        self.send_back = {
            # playback
            'play': bfn[base_addr + '/playback/play'],
            'pause': bfn[base_addr + '/playback/pause'],
            'reverse': bfn[base_addr + '/playback/reverse'],
            'random': bfn[base_addr + '/playback/random'],
            'clear': bfn[base_addr + '/playback/clear'],
            'seek': bfn[base_addr + '/playback/seek'],
            # params
            'speed': bfn[base_addr + '/playback/speed'],
            'sens': bfn['/magi/control{}/sens'.format(self.layer)],
            # cue funs
            'cue_set': bfn[base_addr + '/cue'],
            'cue_clear': bfn[base_addr + '/cue/clear'],
            # loop funs
            'loop_set_a': bfn[base_addr + '/loop/set/a'],
            'loop_set_b': bfn[base_addr + '/loop/set/b'],
            'loop_set_a_cur': bfn[base_addr + '/loop/set/cur/a'],
            'loop_set_b_cur': bfn[base_addr + '/loop/set/cur/b'],
            'loop_on_off': bfn[base_addr + '/loop/set/on_off'],
            'loop_type': bfn[base_addr + '/loop/type'],
            'loop_select': bfn[base_addr + '/loop/select'],
            'loop_move': bfn[base_addr + '/loop/select/move'],
            'loop_clear': bfn[base_addr + '/loop/clear'],
            # scratching
            'scratch_start': bfn['/magi/control{}/start'.format(self.layer)],
            'scratch_do': bfn['/magi/control{}/do'.format(self.layer)],
            'scratch_stop': bfn['/magi/control{}/stop'.format(self.layer)],
        }

        self.speed_var.trace('w', self.gen_update_cmd('speed', self.speed_var))
        self.sens_var.trace('w', self.gen_update_cmd('sens', self.sens_var))

        # colors
        self.no_pad_rows = 1
        if C.NO_Q > 4:
            self.no_pad_rows = C.NO_Q // 4
            if C.NO_Q % 4 != 0:
                self.no_pad_rows += 1

        self.pad_colors = [C.themer.linerp_colors(C.CURRENT_THEME.pad_colors[i], self.no_pad_rows) for i in range(4)]

        # let's setup the gooey
        # it looks like
        # __________________________________________________________
        # |________________________________________________________|
        # |                                       |  <  ||  >  X   |
        # |                                       |  +  spd  sens  |
        # |          [progressbar]                |  -   |    |    |
        # |                                       |  O   +    +    |
        # |                                       | [x]  |    |    |
        # |_______________________________________|________________|
        # |         |         |         |         | [qp/lp switch] |
        # | [qp/lp] |         |         |         | [loop ctrls]   |
        # | ________|_________|_________|_________|                |
        # |         |         |         |         | [loop nxt/prv] |
        # |         |         |         |         | [loop in/out]  |
        # | ________|_________|_________|_________|________________|

        self.setup_main_tk()

    # curry

    def gen_update_cmd(self, key, var):
        fun = self.send_back[key]
        tk_var = var

        def fun_tor(*args):
            fun('', tk_var.get())

        return fun_tor

    def gen_send_cmd(self, key, default_value=1):
        fun = self.send_back[key]
        val = default_value

        def fun_tor(*args):
            # print(key, val)
            fun('', val)

        return fun_tor

    def gen_toggle_cmd(self, key, default_values=[False, True]):
        fun = self.send_back[key]
        toggle_lookup = default_values[:]

        def fun_tor(new_val, *args):
            send_val = toggle_lookup[int(new_val)]  # 0, 1 = False, True
            fun('', send_val)

        return fun_tor

    # send funs

    def activate_pad(self, i=-1):
        if (self.qp_lp_var.get()):  # if lp selected
            fun_to_call = self.send_back['loop_select']
        else:
            fun_to_call = self.send_back['cue_set']
        fun_to_call('', i)

    def delet_pad(self, i=-1):
        if (self.qp_lp_var.get()):  # if lp selected
            fun_to_call = self.send_back['loop_clear']
        else:
            fun_to_call = self.send_back['cue_clear']
        fun_to_call('', i)

    def press_pad(self, i=-1):
        if self.deleting_mode:
            self.delet_pad(i)
        else:
            self.activate_pad(i)

    def press_pad_qp(self, i=-1):
        if self.deleting_mode:
            self.send_back['cue_clear'](i)
        else:
            self.send_back['cue_set'](i)

    def press_pad_lp(self, i=-1):
        if self.deleting_mode:
            self.send_back['loop_clear'](i)
        else:
            self.send_back['loop_select'](i)

    def toggle_pad_mode(self, i=-1):
        self.deleting_mode = not self.deleting_mode
        # maybe update gui somehow
        self.update_loop(self.cur_clip)


    # update dispatch

    def update_directly(self, what, n=-1):
        if what in self.direct_updates:
            self.direct_updates[what](n)

    def update_cur_pos(self, pos):
        self.cur_pos = pos
        self.progressbar.pbar_pos = pos

    def update_clip(self, clip):
        self.cur_clip = clip
        if clip is None:
            self.name_var.set("------")
            self.update_clip_params(clip)
            self.name_label.unbind("<Double-Button-1>")
            return
        self.update_name(clip.name)
        self.update_clip_params(clip)
        self.name_label.bind("<Double-Button-1>", self.change_name_dialog)

    def update_clip_params(self, clip, param=None):
        # print('updating', param)
        if param in self.param_dispatch:
            self.param_dispatch[param](clip)
        elif param is None:
            for fun in self.param_dispatch.values():
                fun(clip)

    def update_name(self, new_name):
        new_text = new_name
        text_meas = []
        for c in new_text:
            if c in C.FONT_WIDTHS:
                text_meas.append(C.FONT_WIDTHS[c])
            else:
                text_meas.append(C.FONT_AVG_WIDTH)

        cumm_text_meas = list(accumulate(text_meas))
        if cumm_text_meas[-1] > self.width - 25:
            to_i = bisect.bisect_left(cumm_text_meas, self.width - 25 - 5 * C.FONT_WIDTHS['.'])
            new_text = new_text[:to_i].strip() + ".."
        self.name_var.set(new_text)

    def update_speed(self, clip):
        if clip is None:
            spd = 0.0
        else:
            spd = clip.params['playback_speed']
        self.speed_var.set(spd)

    def update_sens(self, clip):
        if clip is None:
            sens = 0.0
        else:
            sens = clip.params['control_sens']
        self.sens_var.set(sens)

    def update_cues(self, clip):
        if clip is None:
            for i in range(C.NO_Q):
                but = self.pad_buts[i]
                but.config(state='disabled', relief='groove', background='')

                # unbind
                but.unbind("<ButtonPress-1>")
                but.unbind("<ButtonPress-3>")
            self.progressbar.draw_cue_points()
            return

        cp = clip.params['cue_points']
        self.cue_but_states = [cp[i] is not None for i in range(C.NO_Q)]
        self.progressbar.draw_cue_points(cp, self.cue_but_states)
        self.pad_reconfigure()

    def update_loop(self, clip=None):
        self.update_loop_on(clip)
        self.update_loop_type(clip)

        lp = None

        selected_ind = '-'
        i = -1
        if clip is not None:
            ls = clip.params['loop_selection']
            if ls >= 0:
                selected_ind = str(ls)
                i = ls
            lp = clip.params['loop_points']
            self.loop_but_states = [(lp[i] is not None) and (None not in lp[i][:2]) for i in range(C.NO_LP)]
        else:
            self.loop_but_states = [False for i in range(C.NO_LP)]
        if self.deleting_mode:
            self.loop_selected_text_var.set('DELETE')
            self.lp_selected_label.config(background=C.CURRENT_THEME.delete_bg)
        else:
            self.loop_selected_text_var.set('selected [{}]'.format(selected_ind))
            if i >= 0:
                self.lp_selected_label.config(background=self.pad_colors[i % 4][i // 4])
            else:
                self.lp_selected_label.config(background='')

        self.lp_data = lp
        self.pad_reconfigure()

    def update_loop_on(self, clip=None):
        if clip is None:
            loop_state = False
        else:
            loop_state = clip.params['loop_on']
        if self.loop_on_toggle is not None:
            self.loop_on_toggle.update(loop_state)

        x1, x2 = 0, 1
        if loop_state:
            ls = clip.params['loop_selection']
            if (0 <= ls < C.NO_LP):
                lp = clip.params['loop_points']
                check = lp[ls]
                if check is not None:
                    if check[0] is None:
                        check[0] = 0
                    if check[1] is None:
                        check[1] = 1
                    x1, x2 = check[0], check[1]

        self.progressbar.draw_loop_boundaries(x1, x2)

    def update_loop_type(self, clip=None):
        loop_data = self.backend.loop_get(self.layer)
        if loop_data is None:
            loop_type = False
        else:
            loop_type = (loop_data[1][2] == 'b')

        if self.loop_type_switch is not None:
            self.loop_type_switch.update(loop_type)

    def pad_reconfigure(self, *args):
        if self.cur_clip is None:
            return
        if (self.qp_lp_var.get()):  # if lp selected
            from_what = self.loop_but_states
            self.progressbar.draw_loop_bars(self.lp_data, from_what)
            self.progressbar.exit_cue_mode_binds()
        else:
            from_what = self.cue_but_states
            self.progressbar.draw_loop_bars()
            self.progressbar.cue_mode_only_binds()
        for i, yn in enumerate(from_what):
            but = self.pad_buts[i]
            but.config(state='active')
            but.bind("<ButtonPress-1>", self.pad_but_cmds[i][0])
            but.bind("<ButtonPress-3>", self.pad_but_cmds[i][1])

            if yn:
                r, c = i // 4, i % 4
                but.config(background=self.pad_colors[c][r])
                but.config(relief='raised')
            else:
                but.config(background='')
                but.config(relief='flat')

    def resize(self, new_width, minus_controls=False):
        if minus_controls:
            new_width -= self.bottom_right_frame.winfo_width()

        self.width = new_width

        pad_padding = '{0} 15 {0} 15'.format(new_width // 8 - 4)
        for but in self.pad_buts:
            but.config(padding=pad_padding)

        self.progressbar.resize(new_width)
        self.update_cues(self.cur_clip)
        self.update_loop(self.cur_clip)

    # tk setup

    def format_name(self, wname, i=-1):
        if i > -1:
            wname = wname.format(i)
        wname += '_l{}'.format(self.layer)
        return wname

    def setup_main_tk(self):
        self.root_frame = ttk.Frame(self.root, padding='5 1 5 2')
        self.root_frame.dnd_accept = self.dnd_accept  # for dnd

        self.info_frame = ttk.Frame(self.root_frame, relief='ridge', padding='2')
        self.name_label = ttk.Label(self.info_frame, textvariable=self.name_var,
                                    anchor='center', padding='0 1 0 2')

        left_frame_padding = '2 0 5 0'

        self.top_frame = ttk.Frame(self.root_frame)
        self.progress_frame = ttk.Frame(self.top_frame, padding=left_frame_padding)
        self.top_right_frame = ttk.Frame(self.top_frame)

        self.bottom_frame = ttk.Frame(self.root_frame)
        self.pad_but_frame = ttk.Frame(self.bottom_frame, padding=left_frame_padding)
        self.bottom_right_frame = ttk.Frame(self.bottom_frame)

        # pack it up
        self.root_frame.pack(fill=tk.X, expand=True)

        self.info_frame.pack(side=tk.TOP, fill=tk.X, expand=True)
        self.name_label.pack(expand=True, pady=2, fill=tk.X)

        self.top_frame.pack(side=tk.TOP)
        self.progress_frame.pack(side=tk.LEFT)
        self.top_right_frame.pack(side=tk.LEFT)

        self.bottom_frame.pack(side=tk.TOP)
        self.pad_but_frame.pack(side=tk.LEFT)
        self.bottom_right_frame.pack(side=tk.LEFT)

        # control areas
        self.setup_control_frame_top()
        self.setup_control_frame_bottom()

        # progressbar
        self.setup_progressbar()

        # pads
        self.setup_pads()

    def setup_control_frame_top(self):
        self.control_bottom_frame = ttk.Frame(self.top_right_frame)

        control_slice_pads = '6 0 6 2'
        self.control_sens_frame = ttk.Frame(self.control_bottom_frame, padding=control_slice_pads,
                                            name=self.format_name('ctrl_sens'))
        self.control_spd_frame = ttk.Frame(self.control_bottom_frame, padding=control_slice_pads,
                                           name=self.format_name('pb_speed'))

        self.control_bottom_frame.pack(side=tk.TOP, anchor='w')

        self.control_sens_frame.grid(row=1, column=1, rowspan=4)
        self.control_spd_frame.grid(row=1, column=2, rowspan=4)



        # ctrl buts
        ctrl_but_pad = '12 1 12 1'
        playbut = ttk.Button(self.control_bottom_frame, text=">", width=2, padding=ctrl_but_pad, takefocus=False,
                             command=self.gen_send_cmd('play'), name=self.format_name('pb_play'))
        pausebut = ttk.Button(self.control_bottom_frame, text="||", width=2, padding=ctrl_but_pad, takefocus=False,
                              command=self.gen_send_cmd('pause'), name=self.format_name('pb_pause'))
        rvrsbut = ttk.Button(self.control_bottom_frame, text="<", width=2, padding=ctrl_but_pad, takefocus=False,
                             command=self.gen_send_cmd('reverse'), name=self.format_name('pb_reverse'))
        clearbut = ttk.Button(self.control_bottom_frame, text="X", width=2, padding=ctrl_but_pad, takefocus=False,
                              command=self.gen_send_cmd('clear'), name=self.format_name('pb_clear'))

        for i, but in enumerate([rvrsbut, pausebut, playbut, clearbut]):
            but.grid(row=0, column=i, pady=2, padx=2)

        # zoom buts
        def update_zoom_follow(*args):
            self.progressbar.auto_scroll = self.zoom_follow_var.get()

        self.zoom_follow_var.trace('w', update_zoom_follow)

        zoom_in_but = ttk.Button(self.control_bottom_frame, text="+", width=1, takefocus=False,
                                 command=lambda: self.progressbar.adjust_zoom(1.25))
        zoom_out_but = ttk.Button(self.control_bottom_frame, text="-", width=1, takefocus=False,
                                  command=lambda: self.progressbar.adjust_zoom(.75))
        zoom_reset_but = ttk.Button(self.control_bottom_frame, text="o", width=1, takefocus=False,
                                    command=lambda: self.progressbar.reset_zoom())
        zoom_follow_cb = ttk.Checkbutton(self.control_bottom_frame, width=0,
                                         variable=self.zoom_follow_var, takefocus=False)
        self.zoom_control_buts = [zoom_in_but, zoom_out_but, zoom_reset_but, zoom_follow_cb]
        for i, zcb in enumerate(self.zoom_control_buts):
            zcb.grid(row=(i + 1), column=0, sticky='w')

        spd_sens_vars = [(self.sens_var, self.xtra_sens_var), (self.speed_var, self.xtra_speed_var)]

        def gen_update_trace(v1, v2):
            var1, var2 = v1, v2

            def curry_trace(*args):
                get_got = var1.get()
                new_txt = '{:01.2f}'.format(get_got)
                new_txt = new_txt[:4]
                var2.set(new_txt)

            return curry_trace

        def testVal(test_inp):
            if len(test_inp) > 4:
                return False
            elif len(test_inp) == 0:
                return True
            try:
                float(test_inp)
                return True
            except:
                return False

        for svp in spd_sens_vars:
            t_fun = gen_update_trace(*svp)
            svp[0].trace('w', t_fun)

        def setup_slider(frame, text, var1, var2, style):
            label = ttk.Label(frame, text=text, width=4, relief='groove', borderwidth=2)
            scale = ttk.Scale(frame, from_=10.0, to=0.0, variable=var1,
                              orient=tk.VERTICAL, length=66, style=style)
            scale.bind("<MouseWheel>", lambda e: var1.set(var1.get() + (e.delta / (120 / 0.1))))
            val_entry = ttk.Entry(frame, textvariable=var2, width=4,
                                  validate="key")
            val_entry['validatecommand'] = (val_entry.register(testVal), '%P')
            val_entry.bind('<Return>', lambda e: var1.set(var2.get()))
            val_entry.bind('<Up>', lambda e: var1.set(min(var1.get() + 0.05, 10)))
            val_entry.bind('<Down>', lambda e: var1.set(max(var1.get() - 0.05, 0)))
            label.grid(row=1, column=0)
            scale.grid(row=2, column=0, ipady=2)
            val_entry.grid(row=3, column=0)

        # dont want ultra thicc handles
        s = ttk.Style()
        ss = 'Poop.Vertical.TScale'
        s.configure(ss, sliderlength='17.5')

        setup_slider(self.control_sens_frame, 'sens', *spd_sens_vars[0], ss)
        setup_slider(self.control_spd_frame, 'spd', *spd_sens_vars[1], ss)

        # spd buts
        double_but = ttk.Button(self.control_bottom_frame, text="* 2", width=3, takefocus=False,
                                command=lambda: self.speed_var.set(min(10, 2 * self.speed_var.get())),
                                name=self.format_name('pb_spd_double'))
        half_but = ttk.Button(self.control_bottom_frame, text="/ 2", width=3, takefocus=False,
                              command=lambda: self.speed_var.set(0.5 * self.speed_var.get()),
                              name=self.format_name('pb_spd_halve'))

        double_but.grid(row=2, column=3)
        half_but.grid(row=3, column=3)

    def setup_control_frame_bottom(self):
        self.qp_lp_switch = SwitchButton(self.bottom_right_frame, 'QP', 'LP', padding='2',
                                         name=self.format_name('qp_lp_toggle'))
        self.qp_lp_var = self.qp_lp_switch.bool_var
        self.qp_lp_var.trace('w', self.pad_reconfigure)

        self.direct_updates['qp_lp_toggle'] = self.qp_lp_switch.manual_toggle

        self.qp_lp_switch.but_1.grid(row=0, column=0, sticky='we')
        self.qp_lp_switch.but_2.grid(row=0, column=1, sticky='we')

        self.lp_selected_label = ttk.Label(self.bottom_right_frame, textvariable=self.loop_selected_text_var,
                                           relief='groove', padding='2 4 2 4', anchor='center', borderwidth=2,
                                           name=self.format_name('pb_delet_toggle'))
        self.lp_selected_label.grid(row=0, column=3, columnspan=2, sticky='we')
        self.loop_selected_text_var.set('selected [-]')

        self.loop_on_toggle = ToggleButton(self.bottom_right_frame, 'loop on', 7,
                                           padding='20 4 20 4', name=self.format_name('pb_lp_on_off'))
        self.loop_on_var = self.loop_on_toggle.bool_var
        self.loop_on_toggle.send_cmd = self.gen_toggle_cmd('loop_on_off')
        self.loop_on_toggle.but.grid(row=2, column=0, columnspan=2, sticky='we', pady=2)

        ttk.Frame(self.bottom_right_frame, style='fakesep.TFrame', width=1).grid(row=1, column=0, columnspan=5, sticky='we', pady=4)
        ttk.Frame(self.bottom_right_frame, style='fakesep.TFrame', width=1).grid(row=2, column=2, rowspan=2, sticky='ns', padx=2)
        self.loop_type_switch = SwitchButton(self.bottom_right_frame, 'dflt', 'bnce',
                                             padding='2 4 2 4', name=self.format_name('pb_lp_type'))
        self.loop_type_var = self.loop_type_switch.bool_var
        self.loop_type_switch.send_cmd = self.gen_toggle_cmd('loop_type', ['d', 'b'])
        self.loop_type_switch.but_1.grid(row=2, column=3, sticky='we')
        self.loop_type_switch.but_2.grid(row=2, column=4, sticky='we')

        loop_but_pad = '10 4 10 4'
        loop_in_but = ttk.Button(self.bottom_right_frame, text="in", width=3, padding=loop_but_pad, takefocus=False,
                                 command=self.gen_send_cmd('loop_set_a_cur'), name=self.format_name('pb_lp_set_a'))
        loop_out_but = ttk.Button(self.bottom_right_frame, text="out", width=3, padding=loop_but_pad, takefocus=False,
                                  command=self.gen_send_cmd('loop_set_b_cur'), name=self.format_name('pb_lp_set_b'))

        loop_prev_but = ttk.Button(self.bottom_right_frame, text="\\/", width=2, padding=loop_but_pad, takefocus=False,
                                   command=self.gen_send_cmd('loop_move', -1), name=self.format_name('pb_lp_select_next'))
        loop_next_but = ttk.Button(self.bottom_right_frame, text="/\\", width=2, padding=loop_but_pad, takefocus=False,
                                   command=self.gen_send_cmd('loop_move', +1), name=self.format_name('pb_lp_select_prev'))

        for i, lpb in enumerate([loop_in_but, loop_out_but, loop_prev_but, loop_next_but]):
            c = i + (i > 1)
            lpb.grid(row=3, column=c, sticky='we')

    def setup_progressbar(self):
        self.progressbar = ProgressBar(self.progress_frame, self, self.width, 85)

        self.progressbar.send_funs['seek'] = self.send_back['seek']
        self.progressbar.send_funs['cue'] = self.send_back['cue_set']

        self.progressbar.send_funs['loop_set_a'] = self.send_back['loop_set_a']
        self.progressbar.send_funs['loop_set_b'] = self.send_back['loop_set_b']
        self.progressbar.send_funs['loop_set_a_cur'] = self.send_back['loop_set_a_cur']
        self.progressbar.send_funs['loop_set_b_cur'] = self.send_back['loop_set_b_cur']

        self.progressbar.send_funs['loop_select'] = self.send_back['loop_select']

        def set_cue(i, pos):
            self.backend.set_cue_point(self.layer, i, pos)

        self.progressbar.send_funs['set_cue'] = set_cue

        # scroll scratch
        for k in ['scratch_start', 'scratch_do', 'scratch_stop']:
            self.progressbar.send_funs[k] = self.send_back[k]

        # colors
        self.progressbar.colors['loop_bars'] = self.pad_colors
        self.progressbar.setup_after_color_set()

    def setup_pads(self):
        pad_x = self.width // 8 - 4
        pad_str = '{0} 15 {0} 15'.format(pad_x)

        def gen_but_funs(no):
            i = no

            def activate(*args):
                self.activate_pad(i)

            def deactivate(*args):
                self.delet_pad(i)

            return [activate, deactivate]

        for r in range(self.no_pad_rows):
            for c in range(4):
                i = r * 4 + c
                but = ttk.Label(self.pad_but_frame, text=str(i), borderwidth=4,
                                padding=pad_str, relief='flat', name=self.format_name('pb_pad_{}', i))
                but.grid(row=r, column=c)
                but.config(state='disabled')
                but.bind("<ButtonPress-1>", lambda e: print(e))

                self.pad_buts.append(but)
                self.cue_but_states.append(False)
                self.loop_but_states.append(False)
                self.pad_but_cmds.append(gen_but_funs(i))

    def change_name_dialog(self, *args):
        cur_clip = self.backend.clip_storage.current_clips[self.layer]
        if cur_clip is None:
            return
        new_name = tksimpledialog.askstring("rename clip", '', initialvalue=cur_clip.name)
        if new_name:
            # change name
            self.backend.rename_clip(cur_clip, new_name)  # have to do this to update search properly etc

    # tkdnd stuff
    def dnd_accept(self, source, event):
        return self

    def dnd_enter(self, source, event):
        pass

    def dnd_motion(self, source, event):
        pass

    def dnd_leave(self, source, event):
        pass

    def dnd_commit(self, source, event):
        if source.clip is None:
            return
        self.backend.select_clip(source.clip, self.layer)

    def dnd_end(self, target, event):
        pass


class SwitchButton:
    def __init__(self, frame, text1, text2, min_width=5, padding=None, name=None):
        self.bool_var = tk.BooleanVar()
        self.bool_var.set(False)

        self.send_cmd = None
        if name is not None:
            self.but_1 = ttk.Label(frame, text=text1, borderwidth=4, name=name,
                               width=min_width, anchor='e', style='fakebut.TLabel')
        else:
            self.but_1 = ttk.Label(frame, text=text1, borderwidth=4,
                               width=min_width, anchor='e', style='fakebut.TLabel')
        self.but_1.bind('<Button-1>', lambda e: self.switch(False))
        self.but_2 = ttk.Label(frame, text=text2, borderwidth=4,
                               width=min_width, style='fakebut.TLabel')
        self.but_2.bind('<Button-1>', lambda e: self.switch(True))

        if padding is not None:
            self.but_1.config(padding=padding)
            self.but_2.config(padding=padding)

        self.switch(False)

    def switch(self, new_val):
        if self.send_cmd is not None:
            self.send_cmd(new_val)
        else:
            self.update(new_val)

    def manual_toggle(self, *args):
        self.switch(not bool(self.bool_var.get()))

    def update(self, new_val):
        self.bool_var.set(new_val)
        if (new_val):
            # button 2 now
            self.but_2.config(relief='sunken', state='disabled')
            self.but_1.config(relief='raised', state='')
        else:
            self.but_1.config(relief='sunken', state='disabled')
            self.but_2.config(relief='raised', state='')


class ToggleButton:
    def __init__(self, frame, text, min_width=5, padding=None, name=None):
        self.bool_var = tk.BooleanVar()
        if name is not None:
            self.but = ttk.Label(frame, text=text, borderwidth=4, width=min_width,
                                 style='fakebut.TLabel', name=name)
        else:
            self.but = ttk.Label(frame, text=text, borderwidth=4, width=min_width, style='fakebut.TLabel')
        self.but.bind('<Button-1>', self.toggle)

        self.send_cmd = None

        if padding is not None:
            self.but.config(padding=padding)

    def toggle(self, *args):
        self.switch((not self.bool_var.get()))

    def switch(self, new_val):
        if self.send_cmd is not None:
            self.send_cmd(new_val)
        else:
            self.update(new_val)

    def update(self, new_val):
        self.bool_var.set(new_val)
        if (new_val):
            self.but.config(relief='sunken', state='disabled')
        else:
            self.but.config(relief='raised', state='')


class ProgressBar:
    def __init__(self, root, parent, width=300, height=33):
        self.width, self.height = width, height
        self._drag_data = {'x': 0, 'y': 0, 'item': None, 'label': [], 'type': None}

        self.parent = parent

        self.colors = {
            'bg': 'black',
            'bottom_bar': '#aaa',
            'pbar': 'gray',
            'loop_range': '#333',
            'loop_bar': '#666'
        }

        self.pbar_pos = 0
        self.zoom_factor = 1.0
        self.total_width = width

        self.auto_scroll = False

        self.currently_scratching = False
        self.scratch_job = None

        self.refresh_interval = 100

        # for cue points
        self.qp_lines = [None] * C.NO_Q
        self.qp_labels = [None] * C.NO_Q

        # for loops
        self.loop_bars = [None] * C.NO_LP
        self.loop_bars_data = [None] * C.NO_LP
        self.loop_labels = [None] * C.NO_LP

        # fun dispatch
        self.send_funs = {}

        # tk stuff
        self.root = root
        self.frame = ttk.Frame(self.root)
        self.canvas_frame = ttk.Frame(self.frame)
        self.canvas = tk.Canvas(self.canvas_frame, width=width, height=height + 15,
                                bg=self.colors['bg'], scrollregion=(0, 0, width, height),
                                name=self.parent.format_name('ctrl_bar'))
        self.hbar = ttk.Scrollbar(self.canvas_frame, orient=tk.HORIZONTAL)
        self.hbar.config(command=self.canvas.xview)
        self.canvas.config(xscrollcommand=self.hbar.set)

        self.canvas.pack(anchor=tk.W)
        self.canvas_frame.pack(anchor=tk.W, side=tk.LEFT, expand=tk.YES, fill=tk.BOTH)
        self.hbar.pack(anchor=tk.W, side=tk.BOTTOM, expand=tk.YES, fill=tk.BOTH)
        self.frame.pack(anchor=tk.W, side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

        self.setup_canvas()

        self.root.after(self.refresh_interval, self.update_pbar)

    # # # setup

    def setup_canvas(self):
        w, h = self.width, self.height
        self.canvasbg = self.canvas.create_rectangle(0, 0, w, h,
                                                     fill=self.colors['bg'], tag='bg')
        self.bottombg = self.canvas.create_rectangle(0, h, w, h + 15,
                                                     fill=self.colors['bottom_bar'])

    def setup_after_color_set(self):
        for i in range(C.NO_LP):
            r, c = i // 4, i % 4
            self.loop_bars[i] = self.canvas.create_rectangle(0, 0, 0, 0,
                                                             fill=self.colors['loop_bars'][c][r], tag='loop_bar')
            #                                                activefill=self.colors['?'])
        self.pbar = self.canvas.create_line(0, 0, 0, self.height, fill=self.colors['pbar'], width=3)

        self.outside_loop_rect_l = self.canvas.create_rectangle(0, 0, 0, 0,
                                                                fill=self.colors['loop_range'], stipple='gray50',
                                                                tag='loop_limit')
        self.outside_loop_rect_r = self.canvas.create_rectangle(0, 0, 0, 0,
                                                                fill=self.colors['loop_range'], stipple='gray50',
                                                                tag='loop_limit')
        self.actions_binding()

    def cue_mode_only_binds(self):
        self.canvas.tag_bind("loop_limit", "<B1-Motion>", self.find_mouse)
        self.canvas.tag_bind("loop_limit", "<ButtonRelease-1>", self.find_mouse)

    def exit_cue_mode_binds(self):
        self.canvas.tag_unbind("loop_limit", "<B1-Motion>")
        self.canvas.tag_unbind("loop_limit", "<ButtonRelease-1>")

    def actions_binding(self):
        # seeking
        self.canvas.tag_bind("bg", "<B1-Motion>", self.find_mouse)
        self.canvas.tag_bind("bg", "<ButtonRelease-1>", self.find_mouse)


        # cue points
        self.canvas.tag_bind("qp_line", "<ButtonPress-1>", self.find_nearest)
        self.canvas.tag_bind("qp_label", "<ButtonPress-1>", self.find_nearest)
        self.canvas.tag_bind("qp_line", "<B1-Motion>", self.find_mouse)
        self.canvas.tag_bind("qp_line", "<ButtonPress-3>", self.drag_begin)
        self.canvas.tag_bind("qp_line", "<ButtonRelease-3>", self.drag_end)
        self.canvas.tag_bind("qp_line", "<B3-Motion>", self.drag)

        # looping
        self.canvas.tag_bind("loop_bar", "<ButtonPress-1>", self.find_nearest_loop)
        # self.canvas.tag_bind("loop_label", "<ButtonPress-1>", self.find_nearest_loop)
        self.canvas.bind("<Shift-ButtonPress-2>", self.loop_set_a_cur)
        self.canvas.bind("<Control-ButtonPress-2>", self.loop_set_b_cur)
        self.canvas.tag_bind("loop_limit", "<ButtonPress-3>", self.drag_begin)
        self.canvas.tag_bind("loop_limit", "<ButtonRelease-3>", self.drag_end)
        self.canvas.tag_bind("loop_limit", "<B3-Motion>", self.drag)

        # scratching
        self.canvas.bind("<MouseWheel>", self.scroll_scratch)

    # # # draw helpers

    # the actual bar

    def move_bar(self, x):
        new_x = self.total_width * x
        self.canvas.coords(self.pbar, new_x, 0, new_x, self.height)

    def update_pbar(self):
        self.move_bar(self.pbar_pos)

        # check if need to scroll
        if self.auto_scroll:
            csp = self.hbar.get()
            if self.pbar_pos < csp[0]:
                self.canvas.xview('moveto', max(0, self.pbar_pos - (csp[1] - csp[0])))
            elif self.pbar_pos > csp[1]:
                self.canvas.xview('moveto', self.pbar_pos)

        self.root.after(self.refresh_interval, self.update_pbar)

    # cue points

    def draw_cue_points(self, qp_data=None, qp_on_off=None):
        if qp_data is None:
            for i in range(C.NO_Q):
                self.remove_qp(i)
        else:
            for i, qp in enumerate(qp_data):
                if qp_on_off[i]:
                    self.add_qp(qp, i)
                else:
                    self.remove_qp(i)

    def add_qp(self, x_pos, i):
        x_coord = x_pos * self.total_width
        if self.qp_lines[i] is None:
            r, c = i // 4, i % 4
            self.qp_lines[i] = self.canvas.create_line(x_coord, 0, x_coord, self.height,
                                                       activefill='white', fill='#ccc',
                                                       width=3, dash=(4, ), tags='qp_line')
            labelbox = self.canvas.create_rectangle(x_coord, self.height, x_coord + 15,
                                                    self.height + 15, tags='qp_label',
                                                    fill=self.colors['loop_bars'][c][r])
            labeltext = self.canvas.create_text(x_coord, self.height + 14, anchor=tk.SW,
                                                text=" {}".format(i), fill='black',
                                                activefill='white', justify='center',
                                                tags='qp_label')
            self.qp_labels[i] = [labelbox, labeltext]

        else:  # if qp already exists, move its things
            self.canvas.coords(self.qp_lines[i], x_coord, 0, x_coord, self.height)
            self.canvas.coords(self.qp_labels[i][0], x_coord, self.height,
                               x_coord + 15, self.height + 15)
            self.canvas.coords(self.qp_labels[i][1], x_coord, self.height + 14)

    def remove_qp(self, i):
        if self.qp_lines[i] is None:
            return
        self.canvas.delete(self.qp_lines[i])
        self.qp_lines[i] = None
        if self.qp_labels[i]:
            for label_item in self.qp_labels[i]:
                self.canvas.delete(label_item)
            self.qp_labels[i] = None

    # loop points

    def draw_loop_boundaries(self, x1, x2):
        x1, x2 = x1 * self.total_width, x2 * self.total_width
        self.canvas.coords(self.outside_loop_rect_l, 0, 0, x1, self.height)
        self.canvas.coords(self.outside_loop_rect_r, x2, 0, self.total_width, self.height)

    def draw_loop_bars(self, lp_data=None, lp_on_off=None):
        if lp_data is None:
            for i in range(C.NO_LP):
                self.remove_lp(i)
            return
        for i, lpd in enumerate(lp_data):
            if lp_on_off[i]:
                self.add_lp(i, lpd)
            else:
                self.remove_lp(i)
        top_height, nei = self.do_loop_gravity()
        dy = self.height / top_height

        for i in nei:
            lpd = self.loop_bars_data[i]
            self.canvas.coords(self.loop_bars[i],
                               lpd[0], dy * lpd[1],
                               lpd[2], dy * lpd[3])


    def add_lp(self, i, lpd):
        if None in lpd[:3]:
            self.remove_lp(i)
            return
        x1, x2 = lpd[0] * self.total_width, lpd[1] * self.total_width
        # lpd[2] is loop type .. maybe alternative bar config for this?
        # dy = self.height / C.NO_LP
        # y1 = i * dy
        # y2 = y1 + dy
        self.loop_bars_data[i] = [x1, 0, x2, 1]
        # self.canvas.coords(self.loop_bars[i], x1, y1, x2, y2)

    def remove_lp(self, i):
        self.loop_bars_data[i] = None
        if self.loop_bars[i] is None:
            return
        self.canvas.coords(self.loop_bars[i], 0, 0, 0, 0)
        if self.loop_labels[i] is not None:
            self.canvas.coords(self.loop_labels[i], 0, 0, 0, 0)

    def do_loop_gravity(self):
        lbbd = self.loop_bars_data
        # non empty indices
        nei = [i for i, lpd in enumerate(lbbd) if lpd is not None]

        def check_intersect(i1, i2):
            b0, b1 = lbbd[i1][0], lbbd[i1][2]
            c0, c1 = lbbd[i2][0], lbbd[i2][2]
            if c0 < b0:
                return c1 > b0
            else:
                return c0 < b1

        new_y1 = 0
        for j in range(1, len(nei)):
            # check any of the below loop ranges for intersect
            # next bar must go on top of it
            intersect_heights = [lbbd[nei[k]][3]
                                 if check_intersect(nei[k], nei[j])
                                 else 0
                                 for k in range(0, j)]
            new_y1 = max(intersect_heights)
            lbbd[nei[j]][1] = new_y1
            lbbd[nei[j]][3] = new_y1 + 1
        # return the max height & nei
        return (new_y1 + 1, nei)


    # # # event actions

    def resize(self, new_width):
        self.width = new_width
        self.total_width = new_width * self.zoom_factor
        self.canvas.config(width=self.width, scrollregion=(0, 0, self.width, self.height))
        self.canvas.coords(self.canvasbg, 0, 0, self.width, self.height)
        self.canvas.coords(self.bottombg, 0, self.height, self.width, self.height + 15)

    def adjust_zoom(self, by_factor):
        new_factor = self.zoom_factor * by_factor
        new_factor = max(1.0, new_factor)
        actual_scale = new_factor / self.zoom_factor
        self.canvas.scale(tk.ALL, 0, 0, actual_scale, 1)

        self.total_width = new_factor * self.width
        bbox = (0, 0, self.total_width, self.height)
        self.canvas.configure(scrollregion=bbox)
        self.zoom_factor = new_factor

    def reset_zoom(self):
        self.adjust_zoom(1.0 / self.zoom_factor)

    def find_mouse(self, event):
        # for progress bar to follow mouse
        new_x = self.canvas.canvasx(event.x) / self.total_width
        new_x = max(0, (min(new_x, 1)))
        self.pbar_pos = new_x
        self.move_bar(new_x)
        if 'seek' in self.send_funs:
            self.send_funs['seek']('', new_x)

    # loop funs

    def loop_set_a_cur(self, event):
        if 'loop_set_a_cur' in self.send_funs:
            self.send_funs['loop_set_a_cur']('', True)

    def loop_set_b_cur(self, event):
        if 'loop_set_b_cur' in self.send_funs:
            self.send_funs['loop_set_b_cur']('', True)

    # scratching

    def scroll_scratch(self, event):
        if not self.currently_scratching:
            self.currently_scratching = True
            self.send_funs['scratch_start']('', True)
        dt = event.delta / 12
        self.send_funs['scratch_do']('', dt)
        if self.scratch_job is not None:
            self.root.after_cancel(self.scratch_job)
        self.scratch_job = self.root.after(25, self.stop_scratch)

    def stop_scratch(self):
        self.scratch_job = None
        self.currently_scratching = False
        self.send_funs['scratch_stop']('', True)

    # drag n drop

    def drag_begin(self, event):
        # record the item, its location, any associated labels and what type it is
        item = self.canvas.find_closest(self.canvas.canvasx(event.x), self.canvas.canvasy(event.y), halo=5)[0]
        item_tags = self.canvas.gettags(item)
        if not ('qp_line' or 'loop_limit' in item_tags):
            return
        self._drag_data["item"] = item
        if 'qp_line' in item_tags:
            self._drag_data['label'] = self.qp_labels[self.qp_lines.index(item)]
            self._drag_data['type'] = 'qp'
        else:
            self._drag_data['type'] = 'll'
        self._drag_data['x'] = event.x

    def drag_end(self, event):
        if self._drag_data['item'] is None:
            return
        # just to be safe
        if 'set_cue' in self.send_funs:  # don't do anything until we've set our funs
            new_x = self.canvas.canvasx(event.x)
            if new_x < 0:
                new_x = 0
            elif new_x > self.total_width:
                new_x = self.total_width - 2
            send_x = new_x / self.total_width
            if self._drag_data['type'] == 'qp':
                i = self.qp_lines.index(self._drag_data["item"])
                self.send_funs['set_cue'](i, send_x)
            elif self._drag_data['type'] == 'll':
                if self._drag_data['item'] == self.outside_loop_rect_l:
                    self.send_funs['loop_set_a']('', send_x)
                else:
                    self.send_funs['loop_set_b']('', send_x)

        # reset the drag information
        self._drag_data['item'] = None
        self._drag_data['label'] = []
        self._drag_data['type'] = None
        self._drag_data['x'] = 0

    def drag(self, event):
        # move the object the appropriate amount
        if self._drag_data['item']:
            if self._drag_data['type'] == 'qp':
                # compute how much this object has moved
                delta_x = event.x - self._drag_data['x']
                coord = self.canvas.coords(self._drag_data['item'])
                if coord[0] + delta_x < 0:
                    delta_x = -coord[0]
                elif coord[2] + delta_x > self.total_width:
                    delta_x = self.total_width - coord[2]

                self.canvas.move(self._drag_data['item'], delta_x, 0)  # delta_y)
                for label_item in self._drag_data['label']:
                    self.canvas.move(label_item, delta_x, 0)
            else:
                if self._drag_data['item'] == self.outside_loop_rect_l:
                    self.canvas.coords(self.outside_loop_rect_l,
                                       0, 0, event.x, self.height)
                else:
                    self.canvas.coords(self.outside_loop_rect_r,
                                       event.x, 0, self.total_width, self.height)

        # record the new position
        self._drag_data['x'] = event.x

    def find_nearest(self, event):
        if 'cue' not in self.send_funs:
            return
        item = self.canvas.find_closest(event.x, event.y, halo=5)[0]
        if 'qp_label' in self.canvas.gettags(item):
            item = self.canvas.find_closest(event.x - 10, event.y - 20, halo=5)[0]
        if 'qp_line' in self.canvas.gettags(item):
            i = self.qp_lines.index(item)
        else:
            return
        self.send_funs['cue']('', i)

    def find_nearest_loop(self, event):
        if 'loop_select' not in self.send_funs:
            return
        item = self.canvas.find_closest(event.x, event.y, halo=5)[0]
        if item in self.loop_bars:
            i = self.loop_bars.index(item)
        elif item in self.loop_labels:
            i = self.loop_labels.index(item)
        else:
            return
        self.send_funs['loop_select']('', i)


if __name__ == '__main__':

    rootwin = tk.Tk()
    ttk.Style().theme_use('clam')
    rootwin.title('test_cc')

    # class FakeBackend:
    #     def __init__(self):
    #         pass

    #     def return_fun(self, *args, **kwds):
    #         pass

    #     def __getattr__(self, *args, **kwds):
    #         tor_str = 'call: {}'.format(args[0])
    #         if len(args) > 1:
    #             tor_str += ' args: [{}]'.format(','.join(args[1:]))
    #         if len(kwds) > 0:
    #             tor_str += ' kwds: {}'.format(kwds)
    #         print(tor_str)
    #         return self.return_fun

    # fake_backend = FakeBackend()

    class FakeGUI:
        def __init__(self, backend, root):
            self.root = root
            self.backend = backend
            self.clip_controls = [None] * C.NO_LAYERS
            C.themer.setup(C.CURRENT_THEME, self.root)


        def update_clip(self, layer, clip):
            # print('update?')
            cc = self.clip_controls[layer]
            if cc is not None:
                cc.update_clip(clip)

        def update_clip_params(self, layer, clip, param):
            # dispatch things according to param
            cc = self.clip_controls[layer]
            if cc is not None:
                cc.update_clip_params(clip, param)

        def update_cur_pos(self, layer, pos):
            # pass along the current position
            cc = self.clip_controls[layer]
            if cc is not None:
                cc.update_cur_pos(pos)

        def update_search(self):
            pass

        def update_cols(self, what, ij=None):
            pass

        def update_clip_names(self):
            for i in range(C.NO_LAYERS):
                cc = self.clip_controls[i]
                if cc is not None:
                    cc.update_name(self.backend.clip_storage.current_clips[i].name)

        def restart(self):
            for i in range(C.NO_LAYERS):
                self.update_clip(i, self.backend.clip_storage.current_clips[i])

        def quit(self, *args):
            self.backend.stop()
            self.root.destroy()

    from sol import magi
    test_backend = magi.Magi()
    test_gui = FakeGUI(test_backend, rootwin)
    test_backend.gui = test_gui

    test_cc = ClipControl(rootwin, test_backend, 0)
    test_gui.clip_controls[0] = test_cc

    test_backend.start()

    rootwin.protocol("WM_DELETE_WINDOW", test_gui.quit)
    rootwin.bind("<Control-q>", test_gui.quit)

    test_gui.restart()

    rootwin.mainloop()



import xml.etree.ElementTree as ET
import os

class ResolumeLoader:
    """
    reads .avc and gets the clips out of them
    tested on resolume arena 5 (& 4) may break in the future

    maybe add check to see if proper hack setup is done
    if not offer to do it for them =)
    """
    def __init__(self):
        self.clips = {}
    
    def import_xml(self,xmlfile):
        tree = ET.parse(xmlfile)
        root = tree.getroot()
        
        # general info
        gen_info = [gi for gi in root.iter('generalInfo')][0].attrib
        comp_name = gen_info['name']
        comp_w, comp_h = gen_info['width'], gen_info['height']
        deck_count = gen_info['deckCount']
        if eval(deck_count) > 1:
            print("incorrect setup! please use only 1 deck")

        # lets get to the clips 
        try:
            decks = [deck for deck in \
                     [d for d in root.iter('decks')][0].iter('deck')]
            xml_clips = decks[0].findall('clip')
        except:
            print('no clips found :(')
            return

        for xml_clip in xml_clips:
            # layer/clip are 0 indexed in xml file but osc command
            # is 1 indexed (nice consistency)
            l = eval(xml_clip.get('layerIndex')) + 1
            c = eval(xml_clip.get('trackIndex')) + 1
            command = "/layer{0}/clip{1}/connect".format(l,c)

            name = xml_clip[0][1].get('value')
            f_name = xml_clip[1][0].get('name')
        
            # params
            params = {}
            xml_param = xml_clip[1][1].find('parameters')[0]
            try:
                pbs = eval(xml_param.find('speedFactor').get('curValue'))
            except:
                pbs = 1.0
            params['playback_speed'] = pbs
            direction_lookup = ['b','f','p','r']
            try:
                # fix this... not working >:(
                p_dir_i = eval(xml_param.find('playmode').get('value'))
            except:
                p_dir_i = 2
            params['play_direction'] = direction_lookup[p_dir_i]

            # queue points already present
            # xml_clip[1][1][2][0][0][6] .findall('choice') .get('value')

            # now to figure out how to create clips properly from this
            self.clips[f_name] = (f_name, command, name,None, # none because no thumb yet
                        params) 

    def export_subset(self,xmlfile,output_xml=None,clip_fnames=[]):
        if output_xml is None or len(output_xml)==0:
            reversed_outputname = xmlfile[::-1]
            output_xml = reversed_outputname[reversed_outputname.index('.')+1:][::-1] + '_just_collections.avc'
        # check for proper extension    
        else:
            if output_xml[-4:] != '.avc':
                output_xml += '.avc'
        if os.path.exists(output_xml):
            os.remove(output_xml)

        tree = ET.parse(xmlfile)
        root = tree.getroot()

        decks = [deck for deck in \
                 [d for d in root.iter('decks')][0].iter('deck')]
        for deck in decks:
            xml_clips = deck.findall('clip')
            for clip in xml_clips:
                try:
                    clip_fname = clip[1][0].get('name')
                    if clip_fname not in clip_fnames:
                        deck.remove(clip)
                except Exception as e:
                    pass
        tree.write(output_xml)
        print('successfully exported',output_xml)


if __name__ == '__main__':
    # testfile = "C:/Users/shapil/Documents/Resolume Arena 5/compositions/example.avc"
    testfile = "C:/Users/Leo/Documents/Resolume Arena 5/compositions/vjcomp.avc"


    test_loader = ResolumeLoader()

    test_clip_fnames = ['C:\\VJ\\__clips__\\artsy\\dxv\\Haruhi Spooky Dance.mov', 'C:\\VJ\\__clips__\\artsy\\dxv\\boot tap.mov', 'C:\\VJ\\__clips__\\artsy\\dxv\\Adam Eve.mov', 'C:\\VJ\\__clips__\\artsy\\dxv\\nuke boom.mov', 'C:\\VJ\\__clips__\\artsy\\dxv\\Rainy Un Soldier.mov', 'C:\\VJ\\__clips__\\artsy\\dxv\\sayonara zetsubou sensei moth goodbye.mov', 'C:\\VJ\\__clips__\\artsy\\dxv\\Warning.mov', 'C:\\VJ\\__clips__\\grils\\dxv\\cyberpunk girl thing.mov', 'C:\\VJ\\__clips__\\mecha\\dxv\\Macross Frontier Singing Gril.mov', 'C:\\VJ\\__clips__\\grils\\dxv\\Macross Singing Gril.mov', 'C:\\VJ\\__clips__\\mecha\\dxv\\Macross Dynamite Music Power.mov', 'C:\\VJ\\__clips__\\mecha\\dxv\\Macross Target N Transform.mov', 'C:\\VJ\\__clips__\\mecha\\dxv\\Rock On.mov', 'C:\\VJ\\__clips__\\grils\\dxv\\Photo Frame.mov', 'C:\\VJ\\__clips__\\mecha\\dxv\\Macross Guitar Listen To My Song.mov', 'C:\\VJ\\__clips__\\mecha\\dxv\\Macross Space Station Battle.mov', 'C:\\VJ\\__clips__\\mecha\\dxv\\Space Plane Dodge N Transform.mov', 'C:\\VJ\\__clips__\\mecha\\dxv\\Space Fight Lots Of Mechs Dyin.mov', 'C:\\VJ\\__clips__\\mecha\\dxv\\Space Ship Grils Missiles Etc.mov']

    # test_loader.export_subset(testfile,clip_fnames=test_clip_fnames)
    print(testfile[-4:])




from osc import OscServer
import time

if __name__ == '__main__':
    test_server = OscServer()
    test_server.dispatcher._default_handler = print
    test_server.start()
    while True:
        try:
            time.sleep(1)
        except (KeyboardInterrupt, SystemExit):
            print("exiting...")
            test_server.stop()
            break


from lxml import etree as ET
from xml.dom import minidom
from bisect import bisect_left

import os
import subprocess

from sol.database.clip import Clip
from sol.database.clip import ClipCollection

from sol.config import GlobalConfig
C = GlobalConfig()


class Database:
    """
    hold everything together and has methods to save/load from disk
    so, contains all of our clips, tags, etc
    """

    def __init__(self):
        self.file_ops = FileOPs()
        self.clips = {}
        self.searcher = ClipSearch(self.clips)
        self.tagdb = TagDB()
        self.search = self.searcher.search
        self.default_params = {
        'play_direction' : 'p',
        'playback_speed' : 1.0,
        'cue_points'     : [None] * C.NO_Q,
        'loop_points'    : [None] * C.NO_LP,
        'loop_selection' : -1,
        'loop_on'        : False,
        'duration'       : 0.0,
        'control_sens'   : 1.0
        }

    @property
    def last_search(self):
        return self.searcher.search_res

    @property
    def alphabetical_listing(self):
        all_clips = [(clip.name, clip) for clip in self.clips.values()]
        all_clips.sort()
        return all_clips

    @property
    def hierarchical_listing(self):
        """
        builds out a tree from all clip filenames then flattens it
        and returns folder name / filenames but ignores top-level
        and redundant folders
        i.e. C:\VJ\__clips__\gundam\dxv\g gundam gettin changed.mov
        would just return gundam for folders and full filename..
        """
        if len(self.clips) == 0:
            return
        hierarchy = FileHierarchy()
        all_clips = [(clip.f_name, clip) for clip in self.clips.values()]
        all_clips.sort()
        for clip_n_c in all_clips:
            hierarchy.add_clip(clip_n_c[1])
        tor = []

        def traverse(node):
            # print(node)
            if node.f_or_f == 'clip':
                return [('clip', node.name, node.data)]
            else:
                tor = [('folder', node.name, node.data)]
                for child in node.children:
                    tor += traverse(child)
                return tor

        tor += traverse(hierarchy.root_node)
        return tor
    # def __str__(self,prefix=""):
    #   tor = prefix + self.name + "\n"
    #   for child in self.children:
    #       tor += child.__str__(prefix+"\t")
    #   return tor

    @property
    def all_tags(self):
        all_tags = list(self.tagdb.tags_to_clips.keys())
        all_tags.sort()
        return all_tags

    def add_clip(self, clip):
        self.init_a_clip(clip)
        self.clips[clip.f_name] = clip
        self.searcher.add_clip(clip)
        self.tagdb.add_clip(clip)

    def init_a_clip(self, clip):
        # add default params (if they dont exist)
        for p, p_val in self.default_params.items():
            if p not in clip.params:
                clip.params[p] = p_val
        # create durations, this is pretty quick even with hundreds of clips
        if clip.params['duration'] == 0.0:
            try:
                try:
                    file_info = subprocess.check_output('ffprobe -i "{}" -show_entries format=duration -v quiet -of csv="p=0"'.format(clip.f_name),
                                                        stderr=subprocess.STDOUT, shell=True)
                except subprocess.CalledProcessError as e:
                    file_info = e.output
                clip.params['duration'] = float(file_info.decode('utf-8'))
            except:
                pass

    def add_a_clip(self, clip):
        # for when adding a single clip
        self.add_clip(clip)
        self.searcher.refresh()

    def remove_clip(self, clip):
        self.searcher.remove_clip(clip)
        self.tagdb.remove_clip(clip)
        if clip.f_name in self.clips:
            del self.clips[clip.f_name]

    def rename_clip(self, clip, new_name):
        # gotta remove and readd for searcher to play nice
        self.searcher.remove_clip(clip)
        clip.name = new_name
        self.searcher.add_clip(clip)
        self.searcher.refresh()

    def move_clip(self, clip, new_fname, new_name=None):
        self.remove_clip(clip)
        clip.f_name = new_fname
        if new_name is not None:
            clip.name = new_name
        self.add_a_clip(clip)

    def clear(self):
        self.clips = {}
        self.searcher.clear()
        self.tagdb.clear()


class Search:
    """
    trie-based search interface
    index is of the following form
        (name, reference to thing)
        names do not have to be unique
    can be used for clips
    can be used for tags by making fake clip collections idk
    """

    def __init__(self):
        self.index = []

    def refresh(self):
        # run this after adding many things
        self.index.sort()

    def add_thing(self, name, thing):
        self.index.append((name.lower(), thing))
        for ix, c in enumerate(name):
            if c == " " or c == "_":
                self.index.append((name[ix + 1:].lower(), thing))

    def remove_thing(self, name, thing):
        if (name.lower(), thing) in self.index:
            to_remove = [(name.lower(), thing)]
            for ix, c in enumerate(name.lower()):
                    if c == " " or c == "_":
                        to_remove.append((name[ix + 1:].lower(), thing))
            for rem in to_remove:
                self.index.remove(rem)

    def search_by_prefix(self, prefix, n=-1):
        # return things with names starting with a given prefix
        # in lexicographical order
        tor = set([])
        prefix = prefix.lower()
        i = bisect_left(self.index, (prefix, ''))
        if n < 0:
            till = len(self.index)
        else:
            till = n
        while len(tor) <= till:
            if 0 <= i < len(self.index):
                found = self.index[i]
                if not found[0].startswith(prefix):
                    break
                tor.add(found[1])
                i = i + 1
            else:
                break
        tor = list(tor)
        tor.sort()
        return tor


class ClipSearch(Search):

    def __init__(self, clips):
        super().__init__()
        for clip in clips.values():
            # assuming clips are passed in as a dict
            self.add_clip(clip)
        self.refresh()
        self.search_res = []

    def clear(self):
        self.search_res = []
        self.index = []

    def add_clip(self, clip):
        super().add_thing(clip.name, clip)

    def remove_clip(self, clip):
        super().remove_thing(clip.name, clip)

    def search(self, search_term):
        self.search_res = super().search_by_prefix(search_term)
        return self.search_res

    def refresh(self):
        super().refresh()


class TagDB(Search):

    def __init__(self):
        super().__init__()
        self.clear()

    def add_clip(self, clip):
        for tag in clip.tags:
            self.add_tag(tag, clip)
        self.refresh()

    def add_tag(self, tag, clip):
        if tag not in self.tags_to_clips:
            self.tags_to_clips[tag] = [clip]
            super().add_thing(tag, tag)
            self.needs_refresh = True
        else:
            self.tags_to_clips[tag].append(clip)

    def add_tag_to_clip(self, tag, clip):
        clip.add_tag(tag)
        self.add_tag(tag, clip)
        self.refresh()

    def update_clip_tags(self, tag_list, clip):
        # accepts a list of tag-boolean tuples
        # then either adds or removes said tag from the clip
        for tb in tag_list:
            if len(tb) == 2:
                if tb[1]:
                    self.add_tag_to_clip(tb[0], clip)
                else:
                    self.remove_tag_from_clip(tb[0], clip)
        self.refresh()

    def remove_clip(self, clip):
        for tag in clip.tags:
            self.remove_tag(tag, clip)
        self.refresh()

    def remove_tag(self, tag, clip):
        if tag in self.tags_to_clips:
            if clip in self.tags_to_clips[tag]:
                if len(self.tags_to_clips[tag]) == 1:
                    del self.tags_to_clips[tag]
                    super().remove_thing(tag, tag)
                    self.needs_refresh = True
                else:
                    self.tags_to_clips[tag].remove(clip)
        else:
            super().remove_thing(tag, tag)
            self.needs_refresh = True

    def remove_tag_from_clip(self, tag, clip):
        clip.remove_tag(tag)
        self.remove_tag(tag, clip)
        self.refresh()

    def search(self, search_term):
        self.search_res = super().search_by_prefix(search_term)
        to_return = [(tag, self.tags_to_clips[tag]) for tag in self.search_res if tag in self.tags_to_clips]
        return to_return

    def refresh(self):
        if self.needs_refresh:
            super().refresh()
            self.needs_refresh = False

    def clear(self):
        self.tags_to_clips = {}
        self.needs_refresh = False
        self.search_res = []


class FileOPs:
    """
    save/load to disk
    """

    def __init__(self):
        self.last_save = None
        folder_name = C.SAVEDATA_DIR + os.sep
        last_save_fname = os.path.join(C.SAVEDATA_DIR, 'last_save')
        if not os.path.exists(folder_name):
            os.makedirs(folder_name)
        elif os.path.exists(last_save_fname):
            with open(last_save_fname) as last_save:
                self.last_save = last_save.read()

    def update_last_save(self, filename):
        self.last_save = filename
        last_save_fname = os.path.join(C.SAVEDATA_DIR, 'last_save')
        with open(last_save_fname, 'w') as last_save:
            last_save.write(self.last_save)

    def save_settings(self, settings, name="settings"):
        # save a bunch of things passed in as a dict
        # to a subelement with name = to name : )
        settings_el = ET.Element(name)
        for k, v in settings.items():
            setting = ET.SubElement(settings_el, k)
            if isinstance(v, str):
                v = "'{}'".format(v)
            setting.text = str(v)
        return settings_el

    def load_settings(self, settings_el):
        settings = {}
        for setting in settings_el:
            settings[setting.tag] = eval(setting.text)
        return settings

    def pretty_print(self, el):
        rough_string = ET.tostring(el)
        reparsed = minidom.parseString(rough_string)
        return reparsed.toprettyxml(indent="  ", encoding="utf-8")  # alternatively \t

    def save_clip(self, clip):
        clip_element = ET.Element('clip')
        # filename goes into overarching tag
        clip_element.set('filename', clip.f_name)  # should be unique..
        # thumbnails
        thumbs = ET.SubElement(clip_element, 'thumbnails')
        if clip.t_names is not None:
            thumbs.text = ','.join(clip.t_names)
        # name
        name = ET.SubElement(clip_element, 'name')
        name.text = clip.name
        # activation command
        cmd = ET.SubElement(clip_element, 'activate')
        cmd.text = clip.command
        # tags
        tags = ET.SubElement(clip_element, 'tags')
        tags.text = ','.join(clip.tags)
        # params
        params = self.save_settings(clip.params, 'params')
        clip_element.append(params)

        return clip_element

    def load_clip(self, clip_element):
        # fail gracefully :v)
        if clip_element.get('filename') is None:
            return
        filename = clip_element.get('filename')
        parsed_rep = {}
        for child in clip_element:
            parsed_rep[child.tag] = child
        # build dict out of params
        parsed_params = self.load_settings(parsed_rep['params'])

        fix_things = [('cue_points', C.NO_Q), ('loop_points', C.NO_LP)]
        for k, c in fix_things:
            len_diff = c - len(parsed_params[k])
            if len_diff > 0:
                parsed_params[k].extend([None] * len_diff)

        parsed_params['cue_points']
        parsed_thumbs = parsed_rep['thumbnails'].text
        if parsed_thumbs is None:
            thumbs = None
        else:
            # to avoid issues with commas in filenames..
            split_thumbs = parsed_thumbs.split('.png,')
            thumbs = [thumb + '.png' for thumb in split_thumbs[:-1]] + [split_thumbs[-1]]
        parsed_tags = parsed_rep['tags'].text
        if parsed_tags is None:
            tags = []
        else:
            tags = parsed_tags.split(',')
        clip_tor = Clip(filename, parsed_rep['activate'].text,
                        parsed_rep['name'].text, thumbs,
                        params=parsed_params, tags=tags)
        return clip_tor

    def save_clip_col(self, clip_col):
        # save clip by filename
        # this necessitates loading all clips before looking at clip collections
        col_element = ET.Element('clip_collection')

        col_element.set('n', str(len(clip_col.clips)))
        col_element.set('name', clip_col.name)

        clips = ET.SubElement(col_element, 'clips')
        for clip in clip_col.clips:
            clip_el = ET.SubElement(clips, 'clip')
            if clip is not None:
                clip_el.text = clip.f_name

        return col_element

    def load_clip_col(self, col_element, clips_from_db):
        n = eval(col_element.get('n'))
        name = col_element.get('name')
        # fail nicely
        if n is None or name is None:
            return

        clip_sub_elms = [child for child in col_element[0]]
        clip_names = [c.text for c in clip_sub_elms]

        clip_col_tor = ClipCollection(n, name)
        for i in range(n):
            if clip_names[i] in clips_from_db:
                clip_col_tor[i] = clips_from_db[clip_names[i]]
        return clip_col_tor

    def save_clip_storage(self, clip_store):
        clip_storage_el = ET.Element('clip_storage')
        clip_storage_el.set('current_clip_collection',
                            str(clip_store.cur_clip_col))
        current_clips_el = ET.SubElement(clip_storage_el, 'current_clips')
        current_clips_el.append(self.save_clip_col(clip_store.current_clips))
        clip_cols_el = ET.SubElement(clip_storage_el, 'clip_collections')
        for clip_col in clip_store.clip_cols:
            clip_cols_el.append(self.save_clip_col(clip_col))
        return clip_storage_el

    def load_clip_storage(self, clip_storage_el, clips_from_db):
        cur_clip_col = eval(clip_storage_el.get('current_clip_collection'))
        current_clips = self.load_clip_col(clip_storage_el[0][0], clips_from_db)
        cur_cols = []
        for clip_col_el in clip_storage_el[1]:
            cur_cols.append(self.load_clip_col(clip_col_el, clips_from_db))
        dict_tor = {'cur_clip_col': cur_clip_col,
                    'current_clips': current_clips,
                    'clip_cols': cur_cols
                   }
        return dict_tor

    def save_database(self, database):
        db_el = ET.Element('database')
        for clip in database.clips.values():
            db_el.append(self.save_clip(clip))
        return db_el

    def load_database(self, database_el, database):
        database.clear()
        for clip_el in database_el.findall('clip'):
            new_clip = self.load_clip(clip_el)
            database.add_clip(new_clip)
        database.searcher.refresh()

    def save_midi(self, midi_key_list):
        midi_root = self.create_save('midi')
        midi_key_parts = ['cmd_name', 'osc_addr', 'input_name']
        for mk in midi_key_list:
            new_key = ET.SubElement(midi_root, 'midi_cmd')
            for i, k in enumerate(midi_key_parts):
                new_key.set(k, mk[i])
            # save the midi key
            new_key.set('keys', '/'.join(mk[-1]['keys']))
            new_key.set('type', mk[-1]['type'])
            new_key.set('invert', str(mk[-1]['invert']))
            if mk[-1]['factor'] is not None:
                new_key.set('sens', str(mk[-1]['factor']))

        midi_fname = os.path.join(C.SAVEDATA_DIR, 'midi.xml')
        with open(midi_fname, 'wb') as f:
            f.write(self.pretty_print(midi_root))

    def load_midi(self):
        midi_fname = os.path.join(C.SAVEDATA_DIR, 'midi.xml')
        if not os.path.exists(midi_fname):
            return
        tor = []
        try:
            root = ET.parse(midi_fname).getroot()
            for midi_el in root.findall('midi_cmd'):
                cmd_name = midi_el.get('cmd_name')
                input_name = midi_el.get('input_name')
                factor = midi_el.get('sens')
                try:
                    factor = float(factor)
                except:
                    factor = None
                midi_key = {
                    'keys': midi_el.get('keys').split('/'),
                    'type': midi_el.get('type'),
                    'invert': midi_el.get('invert') == 'True',
                    'factor': factor
                }
                if len(midi_key['keys'][0]) > 0:
                    tor.append([cmd_name, input_name, midi_key])
            return tor
        except Exception as e:
            print('failed to load midi config\n', e)
            return

    def create_save(self, name):
        return ET.Element(name)

    def create_load(self, filename):
        parser = ET.XMLParser(encoding="utf-8", remove_blank_text=True)
        tree = ET.parse(filename, parser=parser)
        root = tree.getroot()
        return root


class FileHierarchyNode:
    def __init__(self,name,what,data=None):
        self.name = name
        self.f_or_f = what
        self.data = data
        self.children = []

    def find_name_in_children(self,search_name):
        to_search = [child.name for child in self.children]
        if search_name not in to_search: return
        tor_i = to_search.index(search_name)
        return self.children[tor_i]

    def __str__(self,prefix=""):
        tor = prefix + self.name + "\n"
        for child in self.children:
            tor += child.__str__(prefix+"\t")
        return tor

class FileHierarchy:
    def __init__(self):
        self.root_node = FileHierarchyNode('File Browser','folder')

    def add_clip(self,clip,cur_node=None,tail=None):
        # start out
        if cur_node is None:
            cur_node = self.root_node
        if tail is None:
            tail = self.splitpath(clip.f_name)
            # allows you to ignore subfolders you don't want cluttering db
            tail = self.check_ignore(tail)
            # if 'dxv' in tail:
            #   del tail[tail.index('dxv')]
        head = tail[0]
        # we've reached the filename (guaranteed to be unique i swear)
        if len(tail) == 1:
            clip_node = FileHierarchyNode(clip.name,'clip',clip.f_name)
            cur_node.children.append(clip_node)
            return
        # traversal
        next_node = cur_node.find_name_in_children(head)
        # when next folder we need doesnt exist
        if next_node is None:
            new_node = FileHierarchyNode(head,'folder',cur_node.name)
            cur_node.children.append(new_node)
            self.add_clip(clip,new_node,tail[1:])
        else:
            self.add_clip(clip,next_node,tail[1:])

    def splitpath(self, path, maxdepth=20):
        (head, tail) = os.path.split(path)
        return self.splitpath(head, maxdepth - 1) + [tail] \
            if maxdepth and head and head != path \
            else [ head or tail ]

    def check_ignore(self,path_list):
        return [p for p in path_list if p not in C.IGNORED_DIRS]

if __name__ == '__main__':
    testdb = Database()
    # test_fnames = ['bazin.mov','test.mov','really cool clip.mov']
    # for fname in test_fnames:
    #   testdb.add_a_clip(Clip(fname,"fake act"))
    # print(testdb.search('c')[0])
    # print(testdb.search('t')[0])
    # print(testdb.search('clip')[0])
    # testdb.clips[test_fnames[0]].f_name = 'hahaha.test'
    # print(testdb.search('bazin')[0])
    # testdb.search('bazin')[0].name = 'not funny anymore'
    # print(testdb.search('bazin')[0])
    # testclip = testdb.clips[test_fnames[1]]

    # test_params = {
    #   'queue_points'   : [0.01,0.51,None,0.76,None,None,None,None],
    #   'loop_points'    : [(0.01,0.51,'d'),(0.6,0.7,'b'),None,None],
    #   'loop_selection' : 0,
    #   'loop_on'        : True,
    #   'playback_speed' : 5.2,
    #   'play_direction' : 'p',
    #   'control_speed'  : 3.33
    # }

    # testclip = Clip("test_clip.mov","/fake/act",params=test_params,
    #   tags = ['test_tag1','tag2','tag2','teststst'])

    # xmlclip = testdb.file_ops.save_clip(testclip)
    # # ET.ElementTree(xmlclip).write('./testclip.xml')
    # rough_string = ET.tostring(xmlclip, 'utf-8')
    # reparsed = minidom.parseString(rough_string)
    # print(reparsed.toprettyxml(indent="\t"))
    # # print(ET.dump(xmlclip))
    # test_testclip = testdb.file_ops.load_clip(xmlclip)
    # print(testclip)
    # print(test_testclip)
    # print([testclip,test_testclip])
    # print(test_testclip.params)

    # test_col = ClipCollection()
    # for i in range(len(test_fnames)):
    #   test_col[i] = testdb.clips[test_fnames[i]]
    # testclipcol = testdb.file_ops.save_clip_col(test_col)
    
    # print(testdb.file_ops.pretty_print(testclipcol))
    # test_parsed_col = testdb.file_ops.load_clip_col(testclipcol,testdb.clips)
    # print(len(test_parsed_col))
    # for i in range(8):
    #   print(test_parsed_col[i])
    # print(test_parsed_col[2] == testdb.clips[test_fnames[2]])
    # for p,v in test_parsed_col[2].params.items():
    #   print(p,v)

    test_fnames = ['C:\VJ\__clips__\gundam\dxv\g gundam gettin changed.mov',
            'C:\VJ\__clips__\gundam\dxv\gundam bad blue guy kills feds.mov',
            'C:\VJ\__clips__\gundam\dxv\gundam beamspam suicides etc.mov',
            'C:\VJ\__clips__\gundum\dxv\gundam big battle.mov']
    test_tags = ['cool','funny','epic','stupid']

    import random

    for fname in test_fnames:
        new_clip = Clip(fname,"fake act")
        random.shuffle(test_tags)
        for t in test_tags[0:random.randint(0,len(test_tags))]:
            new_clip.add_tag(t)
        testdb.add_a_clip(new_clip)
    # test_fh = FileHierarchy()
    # test_listing = testdb.hierarchical_listing
    # print(test_fh.root_node)
    # print(test_listing)

    # print(random.choice(test_tags))
    for fn,clip in testdb.clips.items():
        print(clip,clip.tags)
        last_clip = clip

    print(testdb.tagdb.search(""))
    testdb.tagdb.update_clip_tags([('zzz',True),('cool',False)],last_clip)
    print(last_clip.tags)


class Resolume:
    """
    defines the resolume api
    - has multiple functions that 
        return a tuple with (addr, value) to be sent by osc
            why? can add data validation/casting, idk ""Monads"" man
    - sets up interface that keeps track of variables
        that need to be updated by osc as well
    so for ex: getting current clip position on layer 0
        Resolume.current_clip_pos[0]
            setting curr clip pos on layer 0 to 0.5
        Resolume.set_clip_pos(0,0.5)
            returns ("/layer1/video/position/values",0.5)
    """
    def __init__(self,no_layers):
        self.no_layers = no_layers

        # for updating current clip positions
        self.current_clip_pos = [None] * no_layers
        # index current_clip_pos by layer
        self.clip_pos_addr = [None] * no_layers
        for n in range(no_layers):
            self.clip_pos_addr[n] = "/layer{}/video/position/values".format(n+1)
            
        self.external_looping = False
        # have to do something like this as well
        # 1st make a generator that returns something like this (per layer)
        # def fun_tor(_,msg):
        #   try:
        #       self.current_clip_pos[0] = float(msg)
        #       # print(self) # for debugging
        #   except:
        #       pass
        # return fun_tor
        # for addr in self.clip_pos_addr
        #   self.osc_server.map(addr, generated_fun_tor)

    # might add later (control hax)
    # timeline_fix_addr = '/composition/video/effect1/param{}/values'.format(layer)

    def playback_direction(self,layer,direction):
        addr = "/layer{}/video/position/direction".format(layer+1)
        return (addr, direction)

    def play(self,layer):
        return self.playback_direction(layer,1)
    def pause(self,layer):
        return self.playback_direction(layer,2)
    def reverse(self,layer):
        return self.playback_direction(layer,0)
    def random(self,layer):
        return self.playback_direction(layer,3)

    def set_clip_pos(self,layer,pos):
        addr = '/layer{}/video/position/values'.format(layer+1) 
        return (addr, pos)

    def select_clip(self,layer,clip=None):
        # requires activating clips to be set as
        # "send to active layer"
        # REMEMBER TO CALL clip.command with an arg of 1 afterwards too
        addr = "/layer{}/select".format(layer+1)
        return (addr, 1)

    def clear_clip(self,layer):
        addr = "/layer{}/clear".format(layer+1)
        return (addr, 1)

    def set_playback_speed(self,layer,speed):
        addr = "/layer{}/video/position/speed".format(layer+1)
        return (addr, speed / 10.0) # have to map 0 - 10.0 to 0 - 1.0

    # do I need to keep track of what resolume reports as the playback speed 
    # or just keep of it internally...
    # same goes for playback direction hMM


# defines global variables
import os
import pickle
import importlib

import sol.themer as themer

class SingletonDecorator:
    def __init__(self, klass):
        self.klass = klass
        self.instance = None

    def __call__(self, *args, **kwds):
        if self.instance is None:
            self.instance = self.klass(*args, **kwds)
        return self.instance


@SingletonDecorator
class GlobalConfig:
    def __init__(self, load_defaults=False):
        self.dict = self.__dict__

        self.root_path = os.path.expanduser("~/.sol")
        if not os.path.exists(self.root_path):
            os.mkdir(self.root_path)

        self.default_options = {
            # sol params
            'DEBUG'					: True,
            'SAVEDATA_DIR'			: './savedata',
            'SCROT_DIR'				: './scrot',
            'NO_LAYERS'				: 2,
            'NO_Q'					: 8,
            'NO_LP'					: 8,
            'IGNORED_DIRS'			: ['C:\\', 'dxv', 'VJ'],

            # midi things
            'MIDI_ENABLED'          : True,
            'DEFAULT_SENSITIVITY'   : 0.005,  # control hacks
            'SEPARATE_QP_LP'        : False,
            'SEPARATE_DELETE'       : False,

            # network params
            'OSC_PORT'			    : 7001,
            'MTGUI_ENABLED'		    : False,
            'MTGUI_IP_ADDR'		    : '127.0.0.1',

            # model params
            'MODEL_SELECT'						: 'RESOLUME',
            'MODEL_SELECT_OPTIONS'				: ['RESOLUME', 'MPV', 'ISADORA'],
            'RESOLUME_SAVE_DIR'					: '',
            'SUPPORTED_FILETYPES'				: ['.mov', '.webm', '.mp4', '.avi'],
            'XTERNAL_PLAYER_SELECT'				: 'EXTERNAL_PLAYER',
            'EXTERNAL_PLAYER_SELECT_OPTIONS'	: ['MEMEPV', 'EXTERNAL_PLAYER'],
            'MEMEPV_SCRIPT_PATH'				: 'C:\\code\\vj\\memepv\\test.js',
            'EXTERNAL_PLAYER_COMMAND'			: '',

            # gui params
            'ALWAYS_ON_TOP' 	: True,
            'THUMB_W' 			: 192,
            'REFRESH_INTERVAL'  : 200,
            'FFMPEG_PATH'		: '',
            'NO_FRAMES'			: 5,
            'THUMBNAIL_WIDTH'	: 192,
            'FONT_WIDTHS'		: {},
            'FONT_HEIGHT'		: 10,
            'FONT_AVG_WIDTH'	: 5,
            'SELECTED_THEME'    : 'dark_like_my_soul',
        }

        self.default_options['RESOLUME_SAVE_DIR'] = os.path.expanduser(
            os.sep.join(['~', 'Documents', 'Resolume Arena 5', 'compositions']))
        self.default_options['SAVEDATA_DIR'] = os.path.join(
            self.root_path, 'savedata')
        self.default_options['SCROT_DIR'] = os.path.join(
            self.root_path, 'scrot')

        self.config_savefile = os.path.join(
            self.default_options['SAVEDATA_DIR'], 'config.pickle')

        if load_defaults:
            self.load()
        else:
            self.load(self.config_savefile)

        check_folders = ['SAVEDATA_DIR', 'SCROT_DIR']
        for cf in check_folders:
            folder = self.__dict__[cf]
            if not os.path.exists(folder):
                os.mkdir(folder)

        self.setup_theme()

    def setup_theme(self):
        self.themer = themer
        theme_package_name = 'sol.themes.' + self.dict['SELECTED_THEME']
        self.CURRENT_THEME = importlib.import_module(theme_package_name)

    def save(self):
        del self.themer
        del self.CURRENT_THEME
        try:
            with open(self.config_savefile, 'wb') as save_handle:
                pickle.dump(self.dict, save_handle,
                            protocol=pickle.HIGHEST_PROTOCOL)
            if self.DEBUG:
                print('saved', self.config_savefile)
        except Exception as e:
            print(e)
            if self.DEBUG:
                print('failed to make config savedata')
        self.setup_theme()

    def load(self, load_file=None):
        defaults = self.default_options.items()
        if load_file is not None and os.path.exists(load_file):
            try:
                with open(load_file, 'rb') as save_handle:
                    load_dict = pickle.load(save_handle)
                self.__dict__ = load_dict
            except:
                if self.DEBUG:
                    print('failed to load config savedata\nloading defaults...')

        for k, v in defaults:
            if k not in self.dict:
                self.dict[k] = v



import rtmidi

# import pprint
# pp = pprint.PrettyPrinter(indent=4)

from sol.config import GlobalConfig
C = GlobalConfig()

class LastNList:
    def __init__(self, n, first_el=None):
        self.n = n
        self.list = []
        if first_el is not None:
            self.append(first_el)

    def append(self, what):
        self.list.append(what)
        if len(self.list) > self.n:
            return self.list.pop(0)

    def __repr__(self):
        return self.list.__repr__()


class MidiInterface:
    def __init__(self, magi):
        self.magi = magi
        # midi callback fun references a handle midi cmd
        # that switches btwn config and live
        self.handle_midi = None

        # for configuration
        self.midi_config = None
        self.midi_queue = None
        self.midi_keys_queue = None
        self.midi_config_callback = None

        # things to keep track of
        self.input_ports = {}
        self.midi_inputs = {}
        self.refresh_input_ports()

        # mappings
        self.wname_to_names = {} # when a single widget has multiple associated cmds
        self.all_wnames = []
        self.double_width_pls = [] # relevant for overlay
        self.name_to_cmd = {}

        self.key_to_fun = {} # final mapping of controller -> key -> backend function

        self.gen_name_to_cmd()

    def enable_midi(self):
        if C.MIDI_ENABLED:
            inputs_to_enable = self.map_fun_keys()
            self.handle_midi = self.pass_midi
            for inp in inputs_to_enable:
                self.open_midi_input(inp)
        else:
            self.handle_midi = None

    def enter_config_mode(self, mc):
        self.close_all_inputs()
        self.midi_config = mc
        self.midi_queue = {}
        self.midi_keys_queue = LastNList(5)
        self.handle_midi = self.id_midi
        self.key_to_fun = {}  # remapped on close

    def exit_config_mode(self):
        self.close_all_inputs()
        self.midi_config = None
        self.midi_queue = None
        self.midi_keys_queue = None
        self.enable_midi()

    def id_midi(self, input_name, key, n):
        # used for configuration
        if self.midi_queue is not None:
            mc_exists = self.midi_config is not None
            if key in self.midi_queue:
                self.midi_queue[key].append(n)
            else:
                pos_drop_key = self.midi_keys_queue.append(key)
                if mc_exists:
                    self.midi_config.add_key(key)
                self.midi_queue[key] = LastNList(5, n)
                if pos_drop_key and pos_drop_key in self.midi_queue:
                    del self.midi_queue[pos_drop_key]
                    if mc_exists:
                        self.midi_config.drop_key(pos_drop_key)
        # pp.pprint(self.midi_queue)
        if self.midi_config is not None:
            if self.midi_config_callback is not None:
                self.midi_config.root.after_cancel(self.midi_config_callback)
            self.midi_config_callback = self.midi_config.root.after(100, self.midi_config.update_vals)

    def midi_classify(self):
        if self.midi_queue is None:
            return
        return [(k, self.midi_hist(self.midi_queue[k].list)) for k in self.midi_queue]

    def midi_hist(self, key_list):
        def gen_hist(s, d={}):
            return ([d.__setitem__(i, d.get(i, 0) + 1) for i in s], d)[-1]
        return gen_hist(key_list)

    def pass_midi(self, input_name, key, n=0):
        # used for actual midi control
        if input_name in self.key_to_fun:
            if key in self.key_to_fun[input_name]:
                # print(key, n)
                for fun in self.key_to_fun[input_name][key]:
                    fun(n)

    def refresh_input_ports(self):
        # get a list of all inputs
        self.close_all_inputs()
        self.input_ports = {}
        for api in rtmidi.get_compiled_api():
            try:
                midi = rtmidi.MidiIn(api)
                ports = midi.get_ports()
            except:
                continue
            if ports:
                for num, name in enumerate(ports):
                    self.input_ports[name] = (api, num)
            del midi

    def open_midi_input(self, input_name):
        if input_name not in self.input_ports:
            return
        self.close_midi_input(input_name)
        try:
            port_nos = self.input_ports[input_name]
            new_input = rtmidi.MidiIn(port_nos[0])
            new_input.open_port(port_nos[1])
            new_input.set_callback(self.midi_callback_fun, data=input_name)
            new_input.set_error_callback(self.midi_error_fun)
            self.midi_inputs[input_name] = new_input
        except:
            self.close_midi_input(input_name)

    def close_all_inputs(self):
        for inp in self.input_ports:
            self.close_midi_input(inp)
        if self.midi_queue is not None:
            self.midi_queue = {}
            self.midi_keys_queue = LastNList(5)

    def close_midi_input(self, name):
        if name in self.midi_inputs:
            self.midi_inputs[name].close_port()
            self.midi_inputs[name].cancel_callback()
            self.midi_inputs[name].cancel_error_callback()
            del self.midi_inputs[name]

    def midi_callback_fun(self, midi_tuple, input_name):
        # print(midi_tuple)
        if self.handle_midi is not None:
            key, n = str(midi_tuple[0][:2]), midi_tuple[0][2]
            self.handle_midi(input_name, key, n)

    def osc2midi(self, _, osc_msg):
        msg = eval(osc_msg)
        self.midi_callback_fun(msg, 'osc')

    def midi_error_fun(self, error, error_msg):
        pass

    def gen_name_to_cmd(self):
        # commands that get scaled [default_factor, locked]
        cmds_with_factors = {
            'ctrl_do': [1, False],
            'ctrl_sens': [10, True],
            'pb_spd': [10, True],
            'pb_spd_adjust': [0.1, False]
        }

        cwfk = list(cmds_with_factors.keys())
        for l in range(C.NO_LAYERS):
            for f_k in cwfk:
                cmds_with_factors[f_k + '_l{}'.format(l)] = cmds_with_factors[f_k]

        def add_cmd(cmd_array, l=-1, i=-1):
            if len(cmd_array) < 5:
                cmd_array.extend([None] * (5 - len(cmd_array)))
            [cmd_name, cmd_osc, cmd_desc, wname, combined] = cmd_array
            wn = '{}'
            if wname is not None:
                wn = wname
            if l > -1:
                cmd_name += '_l{}'.format(l)
                if not combined:
                    wn += '_l{}'.format(l)
                cmd_osc = cmd_osc.format(l)
            if i > -1:
                cmd_name = cmd_name.format(i)
                wn = wn.format(i)
                cmd_osc += ' {}'.format(i)
            self.name_to_cmd[cmd_name] = {'addr': cmd_osc, 'desc': cmd_desc, 'factor': None, 'midi_keys': {}}

            if cmd_name in cmds_with_factors:
                self.name_to_cmd[cmd_name]['factor'] = cmds_with_factors[cmd_name]

            if wname is not None:
                if wn in self.wname_to_names:
                    self.wname_to_names[wn].append(cmd_name)
                else:
                    self.wname_to_names[wn] = [cmd_name]
                    self.all_wnames.append(wn)
            else:
                self.all_wnames.append(cmd_name)

        # not per layer cmds
        # cmds are as follows: cmd_name, osc_addr, cmd_desc, widget_name
        not_per_layer_cmds = [
        ['coln_prev', '/magi/cur_col/select_left', 'select collection to the left', 'coln_select'],
        ['coln_next', '/magi/cur_col/select_right', 'select collection to the right', 'coln_select'],
        ]

        # per layer cmds
        # timeline control
        per_layer_cmds = [
        ['ctrl_start', '/magi/control{}/start', 'start control', 'ctrl_bar'],
        ['ctrl_stop', '/magi/control{}/stop', 'stop control', 'ctrl_bar'],
        ['ctrl_do', '/magi/control{}/do', 'do control', 'ctrl_bar'],
        ['ctrl_sens', '/magi/control{}/sens', 'modify control sensitivity'],
        # playback
        ['pb_play', '/magi/layer{}/playback/play', 'play'],
        ['pb_pause', '/magi/layer{}/playback/pause', 'pause'],
        ['pb_reverse', '/magi/layer{}/playback/reverse', 'reverse'],
        ['pb_clear', '/magi/layer{}/playback/clear', 'clear'],
        # speed
        ['pb_spd', '/magi/layer{}/playback/speed', 'set speed', 'pb_speed'],
        ['pb_spd_adjust', '/magi/layer{}/playback/speed/adjust', 'adjust speed', 'pb_speed'],
        ['pb_spd_double', '/magi/layer{}/playback/speed/adjust/factor 2', 'increase speed by a factor of 2'],
        ['pb_spd_halve', '/magi/layer{}/playback/speed/adjust/factor 0.5', 'decrease speed by a factor of 2'],
        # loop points
        ['pb_lp_on_off', '/magi/layer{}/loop/on_off', 'de/activate looping'],
        ['pb_lp_type', '/magi/layer{}/loop/type', 'change loop type'],
        ['pb_lp_select_next', '/magi/layer{}/loop/select/move +1', 'select the next loop range'],
        ['pb_lp_select_prev', '/magi/layer{}/loop/select/move -1', 'select the previous loop range'],
        ['pb_lp_set_a', '/magi/layer{}/loop/set/a', 'set the loop range\'s beginning'],
        ['pb_lp_set_b', '/magi/layer{}/loop/set/b', 'set the loop range\'s end'],
        ]
        # cue points
        if C.SEPARATE_QP_LP:
            if C.SEPARATE_DELETE:
                qp_cmds = [
                # x no qp
                ['pb_qp_{}_act', '/magi/layer{}/cue', 'set/activate cue point', 'pb_pad_{}'],
                ['pb_qp_{}_deact', '/magi/layer{}/cue/clear', 'clear cue point', 'pb_pad_{}'],
                ]
                lp_cmds = [
                # x no lp
                ['pb_lp_{}_act', '/magi/layer{}/loop/select', 'select loop range', 'pb_pad_{}'],
                ['pb_lp_{}_deact', '/magi/layer{}/loop/clear', 'clear the loop range', 'pb_pad_{}'],
                ]
            else:
                qp_cmds = [
                ['pb_qp_{}_press', '/magi/gui/layer{}/qp_pad_press', 'press cue point pad', 'pb_pad_{}'],
                ]
                lp_cmds = [
                ['pb_lp_{}_press', '/magi/gui/layer{}/lp_pad_press', 'press loop range pad', 'pb_pad_{}'],
                ]
                per_layer_cmds.append(
                ['pb_delet_toggle', '/magi/gui/layer{}/pads_toggle', 'pad deletion toggle']
                )
        else:
            if C.SEPARATE_DELETE:
                qp_cmds = [
                ['pb_pad_{}_act', '/magi/gui/layer{}/pad_activate', 'activate pad', 'pb_pad_{}'],
                ['pb_pad_{}_deact', '/magi/gui/layer{}/pad_deactivate', 'deactivate pad', 'pb_pad_{}'],
                ]
                lp_cmds = []
                per_layer_cmds.append(
                ['qp_lp_toggle', '/magi/gui/layer{}/qp_lp_toggle', 'qp/lp toggle']
                )

            else:
                qp_cmds = [
                ['pb_pad_{}_press', '/magi/gui/layer{}/qp_pad_press', 'press cue point pad', 'pb_pad_{}'],
                ]
                lp_cmds = []
                per_layer_cmds.extend([
                ['pb_delet_toggle', '/magi/gui/layer{}/pads_toggle', 'pad deletion toggle'],
                ['qp_lp_toggle', '/magi/gui/layer{}/qp_lp_toggle', 'qp/lp toggle']
                ])
        # collections
        qp_cmds.append(                
                ['coln_select_{}', '/magi/cur_col/select_clip/layer{}', 'select a clip from current collection', 'coln_clip_{}', True],
                )

        # aesthetic fixes for toggle switches
        double_mes = ['pb_lp_type', 'qp_lp_toggle']

        for c in not_per_layer_cmds:
            add_cmd(c)

        for l in range(C.NO_LAYERS):

            for c in per_layer_cmds:
                add_cmd(c, l)

            for c in qp_cmds:
                for i in range(C.NO_Q):
                    add_cmd(c, l, i)
            for c in lp_cmds:
                for i in range(C.NO_LP):
                    add_cmd(c, l, i)

            self.double_width_pls.extend([c + '_l{}'.format(l) for c in double_mes])


        # pp.pprint(self.all_wnames)
        # pp.pprint(self.wname_to_names)

    def gen_savedata(self):
        savedata = []
        for cmd_name, vals in self.name_to_cmd.items():
            if len(vals['midi_keys']) > 0:
                for ctrl_name, midi_key in vals['midi_keys'].items():
                    savedata.append([cmd_name, vals['addr'], ctrl_name, midi_key])

        return savedata

    def load(self, savedata):
        # need to go through savedata
        # any cmd_name that's in name_to_cmd gets updated with filled in 'midi_keys'
        # each line is cmd_name, input_name, midi_key (dict)
        for vals in savedata:
            if vals[0] in self.name_to_cmd:
                self.name_to_cmd[vals[0]]['midi_keys'][vals[1]] = vals[2]

    def gen_fun_wrapper(self, m_type, invert, fun):

        def funtor(n=0):
            fun(n)

        if not invert:  # for readability lol
            # simple on/off
            if m_type == 'togl':
                def funtor(n):
                    if n != 0:
                        fun(1)
            # +/- n at a time (relative control)
            elif m_type == 'knob':
                def funtor(n):
                    if n > 64:
                        n = n - 128
                    fun(n)
            # send new value (absolute control)
            elif m_type == 'sldr':
                def funtor(n):
                    fun(n / 127)  # 0 to 127 translates to 0.0 - 1.0

        else:
            if m_type == 'togl':
                def funtor(n):
                    if n == 0:
                        fun(1)
            elif m_type == 'knob':
                def funtor(n):
                    if n > 64:
                        n = n - 128
                    fun(-1 * n)
            elif m_type == 'sldr':
                def funtor(n):
                    fun(1 - (n / 127))

        return funtor

    def map_fun_keys(self):
        to_enable = []
        for cmd_name, vals in self.name_to_cmd.items():
            osc_addr = vals['addr']
            for ctrl_name, midi_key in vals['midi_keys'].items():
                if ctrl_name not in self.key_to_fun:
                    self.key_to_fun[ctrl_name] = {}
                    to_enable.append(ctrl_name)
                # generate proper fun per type for backend_gen_fun
                backend_gen_fun = self.magi.gen_midi_fun(osc_addr, midi_key['factor'])
                if backend_gen_fun is None:
                    continue
                midi_gen_fun = self.gen_fun_wrapper(midi_key['type'], midi_key['invert'], backend_gen_fun)
                for k in midi_key['keys']:
                    if k in self.key_to_fun[ctrl_name]:
                        self.key_to_fun[ctrl_name][k].append(midi_gen_fun)
                    else:
                        self.key_to_fun[ctrl_name][k] = [midi_gen_fun]
        if 'osc' in to_enable:
            to_enable.remove('osc')
        return to_enable



class MemePV:
    """
    defines the memepv api
    """

    def __init__(self, no_layers):

        self.no_layers = 2
        # for updating current clip positions
        self.current_clip_pos = [None] * no_layers
        # index current_clip_pos by layer
        self.clip_pos_addr = [None] * no_layers

        for n in range(no_layers):
            self.clip_pos_addr[n] = "/{}/time".format(n)
        self.external_looping = True
        self.lt_lu = {'d': 0, 'b': 1}

    def play(self, layer):
        return ('/{}/play'.format(layer), 1)

    def pause(self, layer):
        return ('/{}/pause'.format(layer), 1)

    def reverse(self, layer):
        return ('/{}/reverse'.format(layer), 1)

    def random(self, layer):
        return ('/{}/random'.format(layer), 1)

    def set_clip_pos(self, layer, pos):
        return ('/{}/seek'.format(layer), pos)

    def select_clip(self, layer, clip):
        return ('/{}/load'.format(layer), clip.f_name)

    def clear_clip(self, layer):
        return ('/{}/clear'.format(layer), 1)

    def set_playback_speed(self, layer, speed):
        return ('/{}/speed'.format(layer), speed)

    # looping

    def set_loop_a(self, layer, clip, a=0, b=1):
        if a is None:
            a = 0
        return ('/{}/loop_a'.format(layer), a)

    def set_loop_b(self, layer, clip, a=0, b=1):
        if b is None:
            b = 1
        return ('/{}/loop_b'.format(layer), b)

    def set_loop_type(self, layer, lt):
        if lt in self.lt_lu:
            lt = self.lt_lu[lt]
        else:
            lt = 0
        return ('/{}/loop_type'.format(layer), lt)



import os
import time
import sys

from shutil import copyfile

from sol.magi import Magi
from sol.inputs import osc

"""
backend for gui made in vvvv
hosts magi and sends data to an osc_client
"""

from sol.config import GlobalConfig
C = GlobalConfig()


BASE_ADDR = '/modvj/sol_mt/'


class MTGui:
    """
    can be used w/o printing state and since it doesn't control magi
    can be used simulatenously with another gui :^), just make sure to
    wrap all gui update functions to perform both things in that case
    """

    def __init__(self, magi, ip="127.0.0.1", port=6666):
        self.magi = magi
        self.osc_client = osc.OscClient(ip, port)
        self.update_all()

    @property
    def to_copy_dir(self):
        return os.path.join(C.SCROT_DIR, 'TO_COPY')

    def update_all(self):
        self.update_cols('start')
        for i in range(len(self.magi.clip_storage.current_clips)):
            clip = self.magi.clip_storage.current_clips[i]
            if clip is not None:
                self.update_clip(i, clip)

    def update_loop(self, layer, clip):
        # have to send everything as numbers
        # loop on   - off = 0, on = 1
        # loop_type - 'd' = 0, 'b' = 1
        loop_on = 0
        loop_type = 0
        loop_type_lookup = {'d': 0, 'b': 1}
        if clip is not None:
            # loop on
            loop_on = int(clip.params['loop_on'])
            # get the loop
            loop_select = clip.params['loop_selection']
            if loop_select >= 0:
                cur_loop = clip.params['loop_points'][loop_select]
                # loop type
                if cur_loop is not None:
                    loop_type = loop_type_lookup[cur_loop[2]]
        loop_addr = BASE_ADDR + 'layer{}/loop/'.format(layer)
        self.osc_client.build_n_send(loop_addr + 'on_off', loop_on)
        self.osc_client.build_n_send(loop_addr + 'type', loop_type)
        if clip is not None:
            self.update_lp(layer, clip)
            self.update_cur_loop_range(layer, clip)

    def update_cur_loop_range(self, layer, clip):
        cl = self.magi.loop_get(layer)
        loop_addr = BASE_ADDR + 'layer{}/loop/'.format(layer)
        if cl is not None:
            try:
                cur_range = "{:1.4},{:1.4}".format(cl[1][0], cl[1][1])
            except:
                cur_range = "0,1"
            self.osc_client.build_n_send(loop_addr + 'cur_range', cur_range)
        else:
            self.osc_client.build_n_send(loop_addr + 'cur_range', "0,1")

    def update_qp(self, layer, clip):
        qps = clip.params['cue_points']
        qps_str = ",".join(["1" if qp is not None else "0" for qp in qps])
        qp_addr = BASE_ADDR + 'layer{}/qp'.format(layer)
        self.osc_client.build_n_send(qp_addr, qps_str)

    def update_lp(self, layer, clip):
        lps = clip.params['loop_points']
        lps_str = ",".join(["1" if lp is not None else "0" for lp in lps])
        loop_no = clip.params['loop_selection']
        loop_addr = BASE_ADDR + 'layer{}/loop/'.format(layer)
        self.osc_client.build_n_send(loop_addr + 'lp', lps_str)
        self.osc_client.build_n_send(loop_addr + 'lp_i', loop_no)

    def update_speed(self, layer, clip):
        spd = clip.params['playback_speed']
        spd_addr = BASE_ADDR + 'layer{}/spd'.format(layer)
        self.osc_client.build_n_send(spd_addr, spd)

    def update_sens(self, layer, clip):
        sens = clip.params['control_sens']
        sens_addr = BASE_ADDR + 'layer{}/sens'.format(layer)
        self.osc_client.build_n_send(sens_addr, sens)

    def update_cur_clip_sel(self, layer, clip):
        clip_i_addr = BASE_ADDR + 'layer{}/cur_col_clip_i'.format(layer)
        if clip in self.magi.clip_storage.clip_col:
            clip_i = self.magi.clip_storage.clip_col.index(clip)
            # print(self.magi.clip_storage.current_clips.index(clip)) # sanity
            # check
        else:
            clip_i = -1
        self.osc_client.build_n_send(clip_i_addr, clip_i)

    # magi required funs #
    def update_clip(self, layer, clip):
        if clip is None:
            return
        self.update_loop(layer, clip)
        self.update_speed(layer, clip)
        self.update_sens(layer, clip)
        self.update_qp(layer, clip)
        self.update_cur_clip_sel(layer, clip)

    def update_clip_params(self, layer, clip, param):
        param_dispatch = {
            'cue_points': self.update_qp,
            'loop_points': self.update_loop,
            'loop_on': self.update_loop,
            'loop_type': self.update_loop,
            'loop_selection': self.update_loop,
            'playback_speed': self.update_speed,
            'control_sens': self.update_sens
        }
        if param not in param_dispatch:
            return
        param_dispatch[param](layer, clip)

    def update_cur_pos(self, layer, pos):
        pos_addr = BASE_ADDR + 'layer{}/cur_pos'.format(layer)
        self.osc_client.build_n_send(pos_addr, pos)

    def update_search(self):
        pass

    def update_cols(self, what, ij=None):
        # print('updating cols',what)
        # clip_thumbs
        thumbs_to_send = []
        for i in range(len(self.magi.clip_storage.clip_col.clips)):
            cur_col_clip = self.magi.clip_storage.clip_col.clips[i]
            if cur_col_clip is None or cur_col_clip.t_names is None:
                thumbs_to_send += ["None"]
            else:
                t_name = cur_col_clip.t_names[0]
                fname = os.path.split(t_name)[1][::-1]
                first_part = fname[fname.find("_") + 1:]
                thumb_no = first_part[:first_part.find("_")][::-1]
                thumbs_to_send += [thumb_no]
        to_send = ','.join(thumbs_to_send)
        self.osc_client.build_n_send(BASE_ADDR + 'update_thumbs', to_send)
        for i in range(len(self.magi.clip_storage.current_clips)):
            self.update_cur_clip_sel(
                i, self.magi.clip_storage.current_clips[i])

    def update_clip_names(self):
        pass

    # file operations
    def gen_thumbs_for_me(self):
        # makes sure all thumbnail hashes are unique and renames equivalent
        # thumbs

        def change_hash(clip, old_hash, new_hash):
            # clip.t_names
            to_move = clip.t_names
            clip.t_names = [t_name.replace(
                old_hash, new_hash, 1) for t_name in clip.t_names]
            new_names = [t_name for t_name in clip.t_names]
            for i, lp in enumerate(clip.params['loop_points']):
                if lp is not None:
                    if lp[3] is not None:
                        to_move += [lp[3]]
                        new_names += [lp[3].replace(old_hash, new_hash, 1)]
                        clip.params['loop_points'][i][3] = new_names[-1]
            for i in range(len(to_move)):
                try:
                    os.rename(to_move[i], new_names[i])
                except:
                    pass

        def copy_file(src_file, new_hash):
            # src_file = clip.t_names[0]
            new_file = os.path.join(
                self.to_copy_dir, '{}.png'.format(new_hash))
            try:
                copyfile(src_file, new_file)
            except:
                pass

        if not os.path.exists(self.to_copy_dir):
            os.makedirs(self.to_copy_dir)
        thumbz = {}
        for clip in self.magi.db.clips.values():
            test_str = clip.t_names[0]
            fname = os.path.split(test_str)[1][::-1]
            first_part = fname[fname.find("_") + 1:]
            thumb_no = first_part[:first_part.find("_")][::-1]
            try:
                int(thumb_no)
            except:
                print(thumb_no, 'not a number', '\n\t', test_str)
            if thumb_no in thumbz:
                print('not unique', thumb_no)
                # print(os.path.split(test_str)[0] + '_to_move/' + thumb_no + '.png')
                fix_hash = int(thumb_no)
                counter = 0
                while str(fix_hash) in thumbz:
                    fix_hash += 1
                    counter += 1
                fix_hash = str(fix_hash)
                print('now unique {} (+{})'.format(fix_hash, counter))
                change_hash(clip, thumb_no, fix_hash)
                thumbz[thumb_no] = test_str
            else:
                thumbz[thumb_no] = test_str
            copy_file(test_str, thumb_no)
        print('done copying your thumbnails, find them at')
        print(self.to_copy_dir)

    def print_current_state(self):
        to_print = "*-" * 36 + "*\n" + \
            self.print_cur_clip_info() + "\n" + \
            self.print_a_line() + "\n" + \
            self.print_cur_col() + \
            self.print_a_line() + "\n" + \
            self.print_cols()

        print(to_print)

    def print_cur_clip_info(self):
        name_line = []
        pb_line = []
        pos_line = []
        spd_line = []
        sens_line = []
        loop_line_0 = []
        loop_line_1 = []
        qp_line_0 = []
        qp_line_1 = []
        loop_type_lookup = {'b': 'bnce', 'd': 'dflt', '-': '----'}
        lp_line_0 = []
        lp_line_1 = []
        half_qp = 4
        for i in range(C.NO_LAYERS):
            cur_clip = self.magi.clip_storage.current_clips[i]
            if cur_clip is None:
                name_line += [" -" * 8]
                cur_pb = '-'
                cur_spd = 0
                cur_sens = 0
                loop_on_off = 2
                loop_no = -1
                loop_type = '-'
                qp_line_0 += ["[_]" for j in range(half_qp)]
                qp_line_1 += ["[_]" for j in range(half_qp)]
                lp_line_0 += ["[_]" for j in range(half_qp)]
                lp_line_1 += ["[_]" for j in range(half_qp)]
            else:
                name_line += ["{:<16}".format(cur_clip.name[:16])]
                cur_pb = cur_clip.params['play_direction']
                cur_spd = cur_clip.params['playback_speed']
                cur_sens = cur_clip.params['control_sens']
                # qp #
                qp = cur_clip.params['cue_points']
                # half_qp = len(qp) // 2
                qp_line_0 += ["[x]" if qp[j] is not None else "[_]"
                              for j in range(half_qp)]
                qp_line_1 += ["[x]" if qp[i + half_qp] is not None else "[_]"
                              for j in range(half_qp)]
                # lp #
                lp = cur_clip.params['loop_points']
                loop_on_off = cur_clip.params['loop_on']
                loop_no = cur_clip.params['loop_selection']
                if loop_no >= 0 and lp[loop_no] is not None:
                    loop_type = lp[loop_no][2]
                else:
                    loop_type = '-'
                lp_line_0 += ["[x]" if lp[j] is not None else "[_]"
                              for j in range(half_qp)]
                lp_line_1 += ["[x]" if lp[i + half_qp] is not None else "[_]"
                              for j in range(half_qp)]
            # ["            : {} ".format()]
            pb_line += ["   dir :  {}     ".format(cur_pb)]
            spd_line += ["   spd : {0: 2.2f}  ".format(cur_spd)]
            sens_line += ["  sens : {0: 2.2f}  ".format(cur_sens)]
            loop_line_0 += ["  loop :  {}   ".format(
                ['off', 'on ', ' - '][loop_on_off])]
            loop_line_1 += ["    {0:2d} :  {1}  ".format(
                loop_no, loop_type_lookup[loop_type])]
            cur_pos = self.magi.model.current_clip_pos[i]
            if cur_pos is None:
                cur_pos = 0.00
            pos_line += ["   pos : {0: .3f} ".format(cur_pos)]
        return " | ".join(name_line) +\
            "\n" + " | ".join(pb_line) + \
            "\n" + " | ".join(pos_line) + \
            "\n" + " | ".join(spd_line) + \
            "\n" + " | ".join(sens_line) + \
            "\nQP:" + "".join(qp_line_0[:half_qp]) + "  |   " + \
            "".join(qp_line_0[half_qp:]) + "\n   " + \
            "".join(qp_line_1[:half_qp]) + "  |   " + \
            "".join(qp_line_1[half_qp:]) + \
            "\n" + " | ".join(loop_line_0) + \
            "\n" + " | ".join(loop_line_1) + \
            "\nLP:" + "".join(lp_line_0[:half_qp]) + "  |   " + \
            "".join(lp_line_0[half_qp:]) + "\n   " + \
            "".join(lp_line_1[:half_qp]) + "  |   " + \
            "".join(lp_line_1[half_qp:])

    def print_cur_col(self):
        cur_col_text = []
        for i in range(len(self.magi.clip_storage.clip_col.clips)):
            cur_col_clip = self.magi.clip_storage.clip_col.clips[i]
            if cur_col_clip is None:
                cur_col_text += ["[ ______________ ]"]
            else:
                cur_col_text += ["[{:<16}]".format(cur_col_clip.name[:16])]
        final_string = ""
        for j in range(len(cur_col_text) // 4):
            for i in range(4):
                indx = 4 * j + i
                if indx < len(cur_col_text):
                    final_string += cur_col_text[indx]
            final_string += "\n"
        return final_string

    def print_cols(self):
        names = [c_c.name for c_c in self.magi.clip_storage.clip_cols]
        names[self.magi.clip_storage.cur_clip_col] = "[{}]".format(
            names[self.magi.clip_storage.cur_clip_col])
        return ' | '.join(names)

    def print_a_line(self):
        return "=" * 73


def main():
    gen_thumbs_flag = False
    if len(sys.argv) > 1:
        if sys.argv[1] in ['help', '-h', '-help']:
            print("""this is the standalone multitouch gui for sol
don't use any flags to run sol in commandline mode with a multitouch client
use -g to generate thumbs to copy to your multitouch client
use -h to get this help screen""")
            sys.exit()
        elif sys.argv[1] in ['thumbs', '-g']:
            gen_thumbs_flag = True
        else:
            print('unknown command "{}"'.format(" ".join(sys.argv[1:])))
            sys.exit()

    testit = Magi()
    testit.gui = MTGui(testit)
    if gen_thumbs_flag:
        testit.gui.gen_thumbs_for_me()
    else:
        testit.start()
        while True:
            try:
                time.sleep(1)
                testit.gui.print_current_state()
            except (KeyboardInterrupt, SystemExit):
                print("exiting...")
                testit.stop()
                break

    testit.save_to_file(testit.db.file_ops.last_save)



from setuptools import setup, find_packages

packages = find_packages()

READMEME = """sol is an interactive database of clips

it lets you organize and search for clips plus keep associated parameters, cue points, loop points, thumbnails and tags

it controls playback and adds some missing features to re[sol]ume

MIT licensed

# setup instructions

you will need `ffmpeg` and its associated `ffprobe`

`pip install sol_vj`"""

config = {
    'name': 'sol_vj',
    'description': 'sol - fixes for resolume',
    'long_description': READMEME,
    'author': 'pussinboots',
    'author_email': 'mkaynospam@gmail.com',
    'url': 'https://github.com/pussinboot/sol',
    'license': 'MIT',
    'version': '2.1.5',
    'install_requires': [
        'python-osc>=1.6',
        'lxml>=3.6.0',
        'Pillow>=4.1.0',
        'python-rtmidi >= 1.1.0'
    ],
    'packages': packages,
    'package_data': {
        'sol.gui.tk_gui': ['sample_clip.png', 'sample_thumb.png'],
    },
    'classifiers': [
        "Development Status :: 4 - Beta",
        'Intended Audience :: Developers',
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
        'Programming Language :: Python :: 3',
    ],
    'keywords': 'vj, resolume, video',
    'entry_points': {
        'console_scripts': [
            'sol = sol.test_gui:main',
            'sol_mt = sol.mt_gui:main',
            'sol_cl = sol.magi:main'
        ],
    },
}

setup(**config)



import tkinter as tk
from tkinter import ttk
import tkinter.font as tkFont
import tkinter.filedialog as tkfd
import tkinter.messagebox as tkmb

import os
import string
import pkgutil

from sol.config import GlobalConfig
C = GlobalConfig()

import sol.themes

class SetupGui:
    def __init__(self, rootwin=None, parent=None):
        self.rootwin, self.parent = rootwin, parent
        self.root_frame = tk.Toplevel(takefocus=True)
        self.root_frame.title('setup')
        self.root_frame.focus_force()

        self.menubar = tk.Menu(self.root_frame)
        self.menubar.add_command(
            label="undo all changes", command=self.undo_all)
        self.menubar.add_command(
            label="load defaults", command=self.reset_default)

        self.root_frame.config(menu=self.menubar)

        if self.rootwin is not None:
            self.rootwin.withdraw()
            self.major_changes = []

            def close_fun(*args):
                self.close()
                self.rootwin.destroy()
        else:
            self.parent.root.call('wm', 'attributes', '.', '-topmost', '0')
            # baddies
            self.major_changes = ['NO_LAYERS', 'NO_Q', 'NO_LP', 'MODEL_SELECT',
                                  'OSC_PORT', 'MTGUI_ENABLED', 'MTGUI_IP_ADDR',
                                  'XTERNAL_PLAYER_SELECT', 'SELECTED_THEME']

            def close_fun(*args):
                self.close()
                self.parent.setup_gui = None
                self.parent.toggle_on_top()

        self.root_frame.protocol("WM_DELETE_WINDOW", close_fun)

        # self.generate_font_measurements()

        self.instruction_to_fun = {
            'folder_select': self.add_folder_select,
            'file_select': self.add_file_select,
            'int_choice': self.add_int_choice,
            'bool_choice': self.add_bool_choice,
            'float_choice': self.add_float_choice,
            'list_choice': self.add_list_choice,
            'list_enter': self.add_list_enter,
            'str_enter': self.add_str_enter,
        }

        self.name_to_var = {}
        self.name_to_frame = {}

        self.config_book = ttk.Notebook(self.root_frame)
        self.config_book.pack(expand=True, fill=tk.BOTH)

        # generate theme info
        sol_theme_path = os.path.dirname(sol.themes.__file__)
        self.theme_names = [name for _, name, _ in pkgutil.iter_modules([sol_theme_path])]

        # tabs
        self.reset()
        self.generate_font_measurements()

    def reset(self, to_default=False):

        if to_default:
            C.load()

        cur_no_tabs = len(self.config_book.tabs())
        reopen_tab = -1

        if cur_no_tabs > 0:
            cur_tab_id = self.config_book.select()
            reopen_tab = self.config_book.index(cur_tab_id)
            for _ in range(cur_no_tabs):
                self.config_book.forget(0)

        self.param_tab = None
        self.video_tab = None
        self.gui_tab = None

        tab_padding = '5 5 5 5'

        self.param_tab = ttk.Frame(self.root_frame, padding=tab_padding)
        self.network_tab = ttk.Frame(self.root_frame, padding=tab_padding)
        self.video_tab = ttk.Frame(self.root_frame, padding=tab_padding)
        self.gui_tab = ttk.Frame(self.root_frame, padding=tab_padding)

        for tab_name in [(self.param_tab, 'sol config'),
                         (self.network_tab, 'network config'),
                         (self.video_tab, 'video player'),
                         (self.gui_tab, 'gui config')]:
            self.config_book.add(tab_name[0], text=tab_name[1])

        # construction instructions

        # instruction sets are (instruction, hint text, variable name, [any
        # extra variables])
        param_tab_instr = [
            ('label_frame', 'default folders', '', []),
            ('folder_select', 'savedata directory', 'SAVEDATA_DIR', []),
            ('folder_select', 'screenshot directory', 'SCROT_DIR', []),
            ('label_frame', 'sol parameters', '', []),
            ('int_choice', '# of layers', 'NO_LAYERS', []),
            ('int_choice', '# of cue points', 'NO_Q', []),
            ('int_choice', '# of loop ranges', 'NO_LP', []),
            ('list_enter', 'ignored directories', 'IGNORED_DIRS', []),
            ('bool_choice', 'print debug info', 'DEBUG', []),
            ('label_frame', 'midi options', '', []),
            ('bool_choice', 'midi enabled', 'MIDI_ENABLED', []),
            ('bool_choice', 'separate keys for cue/loop points', 'SEPARATE_QP_LP', []),
            ('bool_choice', 'separate keys for de/activation', 'SEPARATE_DELETE', []),
            ('float_choice', 'default control sensitivity', 'DEFAULT_SENSITIVITY', []),
        ]

        network_tab_instr = [
            ('label_frame', 'server config', '', []),
            ('int_choice', 'osc port', 'OSC_PORT', []),
            ('label_frame', 'multitouch client', '', []),
            ('bool_choice', 'enabled', 'MTGUI_ENABLED', []),
            ('str_enter', 'receiving ip', 'MTGUI_IP_ADDR', [])
        ]

        video_tab_instr = [
            ('label_frame', 'video software config', '', []),
            ('list_choice', 'vj software', 'MODEL_SELECT', C.MODEL_SELECT_OPTIONS),
            ('folder_select', 'composition directory', 'RESOLUME_SAVE_DIR', []),
            ('list_enter', 'supported filetypes', 'SUPPORTED_FILETYPES', []),
            ('label_frame', 'external player config', '', []),
            ('list_choice', 'external player', 'XTERNAL_PLAYER_SELECT',
             C.EXTERNAL_PLAYER_SELECT_OPTIONS),
            ('file_select', 'mpv script', 'MEMEPV_SCRIPT_PATH', []),
            ('str_enter', 'external command', 'EXTERNAL_PLAYER_COMMAND', []),
            ('label_frame', 'ffmpeg options', '', []),
            ('folder_select', 'ffmpeg directory (leave blank if in path)',
             'FFMPEG_PATH', []),
            ('int_choice', '# of thumbnails to generate', 'NO_FRAMES', []),
            ('int_choice', 'thumbnail width', 'THUMBNAIL_WIDTH', []),
        ]

        gui_tab_instr = [
            ('label_frame', 'sol options', '', []),
            ('list_choice', 'theme', 'SELECTED_THEME', self.theme_names),
            ('bool_choice', 'always on top', 'ALWAYS_ON_TOP', []),
            ('label_frame', 'thumbnail options', '', []),
            ('int_choice', 'display width', 'THUMB_W', []),
            ('int_choice', 'hover refresh interval (ms)', 'REFRESH_INTERVAL', []),
        ]

        self.compile_config_page(param_tab_instr, self.param_tab)
        self.compile_config_page(network_tab_instr, self.network_tab)
        self.compile_config_page(video_tab_instr, self.video_tab)
        self.compile_config_page(gui_tab_instr, self.gui_tab)

        self.name_to_var['NO_LP'][0].trace('w', self.loop_lim_check)

        if reopen_tab >= 0:
            self.config_book.select(self.config_book.tabs()[reopen_tab])

        self.root_frame.update_idletasks()
        self.root_frame.after_idle(
            lambda: self.root_frame.minsize(max(500, self.root_frame.winfo_width()),
                                            self.root_frame.winfo_height()))

    def undo_all(self):
        self.reset()

    def reset_default(self):
        self.reset(True)

    def close(self):

        type_to_fun = {
            'int': int,
            'bool': bool,
            'str': str,
            'float': float,
            'list': lambda sl: [s.strip() for s in sl.split(',')]
        }

        any_major = False

        for k, (v_var, v_type) in self.name_to_var.items():
            try:
                new_val = type_to_fun[v_type](v_var.get())
                if k in self.major_changes:
                    if C.dict[k] != new_val:
                        any_major = True
                C.dict[k] = new_val
            except:
                pass

        C.save()
        if any_major:
            tkmb.showwarning(
                '', 'you may have to restart sol\nfor your changes to take effect')
        self.root_frame.destroy()

    def generate_font_measurements(self):
        font = tkFont.Font()
        # height
        C.dict['FONT_HEIGHT'] = font.metrics("linespace")
        # measure font widths
        char_widths = {}
        for c in string.printable:
            char_widths[c] = font.measure(c)

        if 'FONT_WIDTHS' in C.dict:
            for k, v in char_widths.items():
                C.dict['FONT_WIDTHS'][k] = v
        else:
            C.dict['FONT_WIDTHS'] = char_widths

        count, running_sum = 0, 0
        for _, v in C.dict['FONT_WIDTHS'].items():
            count += 1
            running_sum += v
        C.dict['FONT_AVG_WIDTH'] = running_sum / count

    def hide_unhide(self, selection, var_names):
        keys_we_want = []
        for k in self.name_to_frame.keys():
            if '_' in k:
                if any([v in k for v in var_names]):
                    keys_we_want.append(k)

        for k in keys_we_want:
            if selection in k:
                self.name_to_frame[k].pack(
                    side=tk.TOP, expand=False, fill=tk.X, anchor='n')
            else:
                self.name_to_frame[k].pack_forget()

    def compile_config_page(self, instruction_set, parent_tab):
        last_label_frame = None
        starting_optionals = []
        for instruction in instruction_set:
            instr_type, instr_text, instr_varn, instr_extr = instruction

            if instr_type == 'label_frame':
                last_label_frame = self.add_label_frame(instr_text, parent_tab)

            elif instr_type in self.instruction_to_fun:
                starting_choice = None
                if instr_varn in C.dict:
                    starting_choice = C.dict[instr_varn]

                if last_label_frame is not None:
                    new_var, var_type, new_frame = self.instruction_to_fun[instr_type](
                        instr_text, last_label_frame, starting_choice, instr_extr)
                    self.name_to_var[instr_varn] = (new_var, var_type)
                    self.name_to_frame[instr_varn] = new_frame

                    if instr_type == 'list_choice':
                        starting_optionals.append(
                            (starting_choice, instr_extr))
        for sop in starting_optionals:
            # print(sop)
            self.hide_unhide(*sop)

    #######################
    # COMPILER HELPER FUNS

    def add_label_frame(self, frame_name, parent_tab):
        new_label_frame = ttk.LabelFrame(
            parent_tab, text=frame_name, padding='5 5 5 5')
        new_label_frame.pack(side=tk.TOP, expand=False, fill=tk.X, anchor='n')
        return new_label_frame

    def add_choice_row(self, parent_frame, hint_text):
        new_frame = ttk.Frame(parent_frame)
        new_frame.pack(fill=tk.X)
        desc_label = ttk.Label(new_frame, text='{} :'.format(
            hint_text), anchor='w', padding='0 5 0 5')
        desc_label.pack(side=tk.LEFT)
        return new_frame

    def add_folder_select(self, hint_text, parent_frame, starting_choice=None, extra_args=None):
        new_frame = self.add_choice_row(parent_frame, hint_text)
        new_var = tk.StringVar()

        if starting_choice is not None:
            new_var.set(str(starting_choice))

        def change_folder():
            ask_fun = tkfd.askdirectory
            new_folder_path = ask_fun(
                parent=parent_frame, title='select folder', mustexist=True)
            if new_folder_path:
                new_folder_path = os.sep.join(new_folder_path.split('/'))
                new_var.set(new_folder_path)

        dot_but = ttk.Button(new_frame, text='..', command=change_folder, width=1, takefocus=False)
        dot_but.pack(side=tk.RIGHT, anchor='e')

        current_path_label = ttk.Label(
            new_frame, textvar=new_var, anchor='w', relief='sunken')
        current_path_label.pack(side=tk.RIGHT, fill=tk.X,
                                anchor='e', expand=True)

        return new_var, 'str', new_frame

    def add_file_select(self, hint_text, parent_frame, starting_choice=None, extra_args=None):
        new_frame = self.add_choice_row(parent_frame, hint_text)
        new_var = tk.StringVar()

        if starting_choice is not None:
            new_var.set(str(starting_choice))

        def change_file():
            ask_fun = tkfd.askopenfilename
            new_file_path = ask_fun(parent=parent_frame, title='select file')
            if new_file_path:
                new_file_path = os.sep.join(new_file_path.split('/'))
                new_var.set(new_file_path)

        dot_but = ttk.Button(new_frame, text='..', command=change_file, width=1, takefocus=False)
        dot_but.pack(side=tk.RIGHT, anchor='e')

        current_path_label = ttk.Label(
            new_frame, textvar=new_var, anchor='w', relief='sunken')
        current_path_label.pack(side=tk.RIGHT, fill=tk.X,
                                anchor='e', expand=True)

        return new_var, 'str', new_frame

    def add_int_choice(self, hint_text, parent_frame, starting_choice=None, extra_args=None):
        new_frame = self.add_choice_row(parent_frame, hint_text)
        new_var = tk.StringVar()

        if starting_choice is not None:
            new_var.set(str(starting_choice))

        no_entry = tk.Spinbox(new_frame, from_=0, to=9999,
                              textvariable=new_var, justify='left', width=5)
        no_entry.pack(side=tk.RIGHT, anchor='e')

        return new_var, 'int', new_frame

    def add_float_choice(self, hint_text, parent_frame, starting_choice=None, extra_args=None):
        new_frame = self.add_choice_row(parent_frame, hint_text)
        new_var = tk.StringVar()

        if starting_choice is not None:
            new_var.set(str(starting_choice))

        no_entry = tk.Spinbox(new_frame, from_=0, to=2, increment=0.005,
                              textvariable=new_var, justify=tk.RIGHT, width=5)
        no_entry.pack(side=tk.RIGHT, anchor='e')

        return new_var, 'float', new_frame

    def add_bool_choice(self, hint_text, parent_frame, starting_choice=None, extra_args=None):
        new_frame = self.add_choice_row(parent_frame, hint_text)
        new_var = tk.IntVar()

        if starting_choice is not None:
            new_var.set(int(starting_choice))

        check_but = ttk.Checkbutton(new_frame, variable=new_var, takefocus=False)
        check_but.pack(side=tk.RIGHT, anchor='e')

        return new_var, 'bool', new_frame

    def add_list_choice(self, hint_text, parent_frame, starting_choice=None, extra_args=None):
        new_frame = self.add_choice_row(parent_frame, hint_text)
        new_var = tk.StringVar()
        pos_values = ' '.join(extra_args)
        selector = ttk.Combobox(new_frame, textvariable=new_var, values=pos_values)
        selector.config(state='readonly')
        selector.pack(side=tk.RIGHT, anchor='e')

        if starting_choice is not None:
            new_var.set(str(starting_choice))

        def gen_hide_callback():
            dis_var = new_var
            x_args = extra_args

            def callback(*args):
                self.hide_unhide(dis_var.get(), x_args)
            return callback

        hide_cb = gen_hide_callback()
        new_var.trace("w", hide_cb)

        return new_var, 'str', new_frame

    def add_list_enter(self, hint_text, parent_frame, starting_choice=None, extra_args=None):
        new_frame = self.add_choice_row(parent_frame, hint_text)
        new_var = tk.StringVar()
        list_entry = ttk.Entry(new_frame, textvariable=new_var, justify="left")
        list_entry.pack(side=tk.RIGHT, fill=tk.X, anchor='e', expand=True)

        if starting_choice is not None:
            starting_text = ", ".join(starting_choice)
            new_var.set(starting_text)

        return new_var, 'list', new_frame

    def add_str_enter(self, hint_text, parent_frame, starting_choice=None, extra_args=None):
        new_frame = self.add_choice_row(parent_frame, hint_text)
        new_var = tk.StringVar()

        if starting_choice is not None:
            new_var.set(str(starting_choice))

        str_entry = ttk.Entry(new_frame, textvariable=new_var, justify="left")
        str_entry.pack(side=tk.RIGHT, fill=tk.X, anchor='e', expand=True)

        return new_var, 'str', new_frame

    ################
    # EXTRA CHECKS

    def loop_lim_check(self, *args):
        try:
            no_lp, no_qp = int(self.name_to_var['NO_LP'][0].get()), int(self.name_to_var['NO_Q'][0].get())
            if no_lp > no_qp:
                self.name_to_var['NO_LP'][0].set(no_qp)
        except:
            pass


if __name__ == '__main__':
    root = tk.Tk()
    root.title('sol')
    SetupGui(root)
    root.mainloop()



from pythonosc import dispatcher
from pythonosc import osc_server
from pythonosc import osc_message_builder
from pythonosc import udp_client

import threading
import re
import os
import socket

"""
provide interfaces for osc server/clients
"""

class OscServer:
    """
    runs an osc server with an event dispatcher on the network
    and on the local machine : )
    """

    def __init__(self,ip=None,port=7001):
        if ip is None:
            ip = self.get_ip_addr()
        self.ip, self.port = ip, port
        self.running = 0
        self.dispatcher = dispatcher.Dispatcher()
        self.map = self.dispatcher.map
        self.number_regex = re.compile(r'(\d+)$')

    def start(self):
        self.running = 1
        print("starting osc server",self.ip,':',self.port,'\n\tand on localhost :',self.port)
        local_ip = '127.0.0.1'
        self.server = osc_server.ThreadingOSCUDPServer((self.ip, self.port), self.dispatcher)
        self.server_thread = threading.Thread(target=self.server.serve_forever)
        self.server_thread.start()

        if self.ip != local_ip:
            self.local_server = osc_server.ThreadingOSCUDPServer((local_ip, self.port), self.dispatcher)
            self.local_server_thread = threading.Thread(target=self.local_server.serve_forever)
            self.local_server_thread.start()
        else:
            self.local_server = None

    def stop(self):
        self.running = 0
        self.server.shutdown()
        self.server_thread.join()
        if self.local_server is not None:
            self.local_server.shutdown()
            self.local_server_thread.join()

    def osc_value(self,msg):
        # returns the value of the osc msg
        tor = msg
        if isinstance(msg,str):
            try:
                tor = eval(msg)
            except:
                pass
        return tor

    def find_num_in_addr(self,addr):
        try:
            return int(self.number_regex.search(addr).group(1))
        except:
            return -1

    def map_unique(self,addr,fun):
        if addr in self.dispatcher._map:
            del self.dispatcher._map[addr]
        self.map(addr,fun)

    def get_ip_addr(self):
        #  from https://stackoverflow.com/questions/166506/finding-local-ip-addresses-using-pythons-stdlib/
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            # doesn't even have to be reachable
            s.connect(('10.255.255.255', 1))
            IP = s.getsockname()[0]
        except:
            IP = '127.0.0.1'
        finally:
            s.close()
        return IP


class OscClient:
    def __init__(self,ip="127.0.0.1",port=7000):
        self.ip, self.port = ip, port
        self.osc_client = udp_client.UDPClient(ip, port)
        self.send = self.osc_client.send

    def build_msg(self,addr,arg):
        msg = osc_message_builder.OscMessageBuilder(address = addr)
        msg.add_arg(arg)
        msg = msg.build()
        return msg

    def build_n_send(self,addr,arg):
        msg = self.build_msg(addr,arg)
        self.osc_client.send(msg)
        return msg

    def build_n_send_bundle(self,addr,argz):
        msg = osc_message_builder.OscMessageBuilder(address = addr)
        for arg in argz:
            msg.add_arg(arg)
        msg = msg.build()
        self.osc_client.send(msg)
        return msg